# -*- coding: utf-8 -*-
# (c) 2020 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import datetime
import locale
import sys
import typing as t
import unicodedata

from unittest.mock import MagicMock

import pytest

from ansible.module_utils.datatag import deprecator_from_collection_name
from ansible.module_utils._internal import _deprecator, _errors, _messages
from ansible.utils.display import _LIBC, _MAX_INT, Display, get_text_width
from ansible.utils.multiprocessing import context as multiprocessing_context
from ansible._internal import _display_utils


@pytest.fixture
def problematic_wcswidth_chars():
    locale.setlocale(locale.LC_ALL, 'C.UTF-8')

    candidates = set(chr(c) for c in range(sys.maxunicode) if unicodedata.category(chr(c)) == 'Cf')
    problematic = [candidate for candidate in candidates if _LIBC.wcswidth(candidate, _MAX_INT) == -1]

    if not problematic:
        # Newer distributions (Ubuntu 22.04, Fedora 38) include a libc which does not report problematic characters.
        pytest.skip("no problematic wcswidth chars found")  # pragma: nocover

    return problematic


def test_get_text_width():
    locale.setlocale(locale.LC_ALL, '')
    assert get_text_width(u'コンニチハ') == 10
    assert get_text_width(u'abコcd') == 6
    assert get_text_width(u'café') == 4
    assert get_text_width(u'four') == 4
    assert get_text_width(u'\u001B') == 0
    assert get_text_width(u'ab\u0000') == 2
    assert get_text_width(u'abコ\u0000') == 4
    assert get_text_width(u'🚀🐮') == 4
    assert get_text_width(u'\x08') == 0
    assert get_text_width(u'\x08\x08') == 0
    assert get_text_width(u'ab\x08cd') == 3
    assert get_text_width(u'ab\x1bcd') == 3
    assert get_text_width(u'ab\x7fcd') == 3
    assert get_text_width(u'ab\x94cd') == 3

    pytest.raises(TypeError, get_text_width, 1)
    pytest.raises(TypeError, get_text_width, b'four')


def test_get_text_width_no_locale(problematic_wcswidth_chars):
    pytest.raises(EnvironmentError, get_text_width, problematic_wcswidth_chars[0])


def test_Display_banner_get_text_width(monkeypatch, display_resource):
    locale.setlocale(locale.LC_ALL, '')
    display = Display()
    display_mock = MagicMock()
    monkeypatch.setattr(display, 'display', display_mock)

    display.banner(u'🚀🐮', color=False, cows=False)
    args, kwargs = display_mock.call_args
    msg = args[0]
    stars = u' %s' % (75 * u'*')
    assert msg.endswith(stars)


def test_Display_banner_get_text_width_fallback(monkeypatch, display_resource):
    locale.setlocale(locale.LC_ALL, 'C.UTF-8')
    display = Display()
    display_mock = MagicMock()
    monkeypatch.setattr(display, 'display', display_mock)

    display.banner(u'\U000110cd', color=False, cows=False)
    args, kwargs = display_mock.call_args
    msg = args[0]
    stars = u' %s' % (78 * u'*')
    assert msg.endswith(stars)


def test_Display_set_queue_parent(display_resource):
    display = Display()
    pytest.raises(RuntimeError, display.set_queue, 'foo')


def test_Display_set_queue_fork(display_resource):
    def test():
        display = Display()
        display.set_queue('foo')
        assert display._final_q == 'foo'
    p = multiprocessing_context.Process(target=test)
    p.start()
    p.join()
    assert p.exitcode == 0


def test_Display_display_fork(display_resource):
    def test():
        queue = MagicMock()
        display = Display()
        display.set_queue(queue)
        display.display('foo')
        queue.send_display.assert_called_once_with('display', 'foo')

    p = multiprocessing_context.Process(target=test)
    p.start()
    p.join()
    assert p.exitcode == 0


def test_Display_display_warn_fork(display_resource):
    def test():
        queue = MagicMock()
        display = Display()
        display.set_queue(queue)
        display.warning('foo')
        queue.send_display.assert_called_once_with('_warning', _messages.WarningSummary(event=_messages.Event(msg='foo')))

    p = multiprocessing_context.Process(target=test)
    p.start()
    p.join()
    assert p.exitcode == 0


def test_Display_display_lock(monkeypatch, display_resource):
    lock = MagicMock()
    display = Display()
    monkeypatch.setattr(display, '_lock', lock)
    display.display('foo')
    lock.__enter__.assert_called_once_with()


def test_Display_display_lock_fork(monkeypatch, display_resource):
    lock = MagicMock()
    display = Display()
    monkeypatch.setattr(display, '_lock', lock)
    monkeypatch.setattr(display, '_final_q', MagicMock())
    display.display('foo')
    lock.__enter__.assert_not_called()


def test_format_message_deprecation_with_multiple_details() -> None:
    """
    Verify that a DeprecationSummary with multiple Detail entries can be formatted.
    No existing code generates deprecations with multiple details, but a future deprecation exception type would need to make use of this.
    """
    event = _messages.Event(
        msg='Ignoring ExceptionX.',
        help_text='Plugins must handle it internally.',
        chain=_messages.EventChain(
            msg_reason=_errors.MSG_REASON_DIRECT_CAUSE,
            traceback_reason='',  # not used in this test
            event=_messages.Event(
                msg='Something went wrong.',
                formatted_source_context='Origin: /some/path\n\n...',
            ),
        ),
    )

    result = _display_utils.format_message(_messages.DeprecationSummary(event=event), False)

    assert result == '''Ignoring ExceptionX. This feature will be removed in the future: Something went wrong.

Ignoring ExceptionX. This feature will be removed in the future. Plugins must handle it internally.

<<< caused by >>>

Something went wrong.
Origin: /some/path

...

'''


A_DATE = datetime.date(2025, 1, 1)
CORE = deprecator_from_collection_name('ansible.builtin')
CORE_MODULE = _messages.PluginInfo(resolved_name='ansible.builtin.ping', type=_messages.PluginType.MODULE)
CORE_PLUGIN = _messages.PluginInfo(resolved_name='ansible.builtin.debug', type=_messages.PluginType.ACTION)
COLL = deprecator_from_collection_name('ns.col')
COLL_MODULE = _messages.PluginInfo(resolved_name='ns.col.ping', type=_messages.PluginType.MODULE)
COLL_PLUGIN = _messages.PluginInfo(resolved_name='ns.col.debug', type=_messages.PluginType.ACTION)
INDETERMINATE = _deprecator.INDETERMINATE_DEPRECATOR
LEGACY_MODULE = _messages.PluginInfo(resolved_name='ping', type=_messages.PluginType.MODULE)
LEGACY_PLUGIN = _messages.PluginInfo(resolved_name='debug', type=_messages.PluginType.ACTION)


@pytest.mark.parametrize('kwargs, expected', (
    # removed
    (dict(msg="Hi", removed=True), "Hi. This feature was removed."),
    (dict(msg="Hi", version="2.99", deprecator=CORE, removed=True), "Hi. This feature was removed from ansible-core version 2.99."),
    (dict(msg="Hi", date=A_DATE, deprecator=COLL_MODULE, removed=True),
     "Hi. This feature was removed from module 'ping' in collection 'ns.col' in a release after 2025-01-01."),
    # no deprecator or indeterminate
    (dict(msg="Hi"), "Hi. This feature will be removed in the future."),
    (dict(msg="Hi", version="2.99"), "Hi. This feature will be removed in the future."),
    (dict(msg="Hi", date=A_DATE), "Hi. This feature will be removed in the future."),
    (dict(msg="Hi", version="2.99", deprecator=INDETERMINATE), "Hi. This feature will be removed in the future."),
    (dict(msg="Hi", date=A_DATE, deprecator=INDETERMINATE), "Hi. This feature will be removed in the future."),
    # deprecator without plugin
    (dict(msg="Hi", deprecator=CORE), "Hi. This feature will be removed from ansible-core in a future release."),
    (dict(msg="Hi", deprecator=COLL), "Hi. This feature will be removed from collection 'ns.col' in a future release."),
    (dict(msg="Hi", version="2.99", deprecator=CORE), "Hi. This feature will be removed from ansible-core version 2.99."),
    (dict(msg="Hi", version="2.99", deprecator=COLL), "Hi. This feature will be removed from collection 'ns.col' version 2.99."),
    (dict(msg="Hi", date=A_DATE, deprecator=COLL), "Hi. This feature will be removed from collection 'ns.col' in a release after 2025-01-01."),
    # deprecator with module
    (dict(msg="Hi", deprecator=CORE_MODULE), "Hi. This feature will be removed from module 'ping' in ansible-core in a future release."),
    (dict(msg="Hi", deprecator=COLL_MODULE), "Hi. This feature will be removed from module 'ping' in collection 'ns.col' in a future release."),
    (dict(msg="Hi", deprecator=LEGACY_MODULE), "Hi. This feature will be removed from module 'ping' in the future."),
    (dict(msg="Hi", version="2.99", deprecator=CORE_MODULE), "Hi. This feature will be removed from module 'ping' in ansible-core version 2.99."),
    (dict(msg="Hi", version="2.99", deprecator=COLL_MODULE), "Hi. This feature will be removed from module 'ping' in collection 'ns.col' version 2.99."),
    (dict(msg="Hi", version="2.99", deprecator=LEGACY_MODULE), "Hi. This feature will be removed from module 'ping' in the future."),
    (dict(msg="Hi", date=A_DATE, deprecator=COLL_MODULE),
     "Hi. This feature will be removed from module 'ping' in collection 'ns.col' in a release after 2025-01-01."),
    (dict(msg="Hi", date=A_DATE, deprecator=LEGACY_MODULE), "Hi. This feature will be removed from module 'ping' in the future."),
    # deprecator with plugin
    (dict(msg="Hi", deprecator=CORE_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in ansible-core in a future release."),
    (dict(msg="Hi", deprecator=COLL_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in collection 'ns.col' in a future release."),
    (dict(msg="Hi", deprecator=LEGACY_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in the future."),
    (dict(msg="Hi", version="2.99", deprecator=CORE_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in ansible-core version 2.99."),
    (dict(msg="Hi", version="2.99", deprecator=COLL_PLUGIN),
     "Hi. This feature will be removed from action plugin 'debug' in collection 'ns.col' version 2.99."),
    (dict(msg="Hi", version="2.99", deprecator=LEGACY_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in the future."),
    (dict(msg="Hi", date=A_DATE, deprecator=COLL_PLUGIN),
     "Hi. This feature will be removed from action plugin 'debug' in collection 'ns.col' in a release after 2025-01-01."),
    (dict(msg="Hi", date=A_DATE, deprecator=LEGACY_PLUGIN), "Hi. This feature will be removed from action plugin 'debug' in the future."),
))
def test_get_deprecation_message_with_plugin_info(kwargs: dict[str, t.Any], expected: str) -> None:
    for kwarg in ('version', 'date', 'deprecator'):
        kwargs.setdefault(kwarg, None)

    msg = _display_utils.get_deprecation_message_with_plugin_info(**kwargs)

    assert msg == expected


@pytest.mark.parametrize("kw,expected", (
    (dict(msg="hi"), "[DEPRECATION WARNING]: hi. This feature will be removed in the future."),
    (dict(msg="hi", removed=True), "[DEPRECATED]: hi. This feature was removed."),
    (dict(msg="hi", version="1.23"), "[DEPRECATION WARNING]: hi. This feature will be removed in the future."),
    (dict(msg="hi", date="2025-01-01"), "[DEPRECATION WARNING]: hi. This feature will be removed in the future."),
    (dict(msg="hi", collection_name="foo.bar"), "[DEPRECATION WARNING]: hi. This feature will be removed from collection 'foo.bar' in a future release."),
    (dict(msg="hi", version="1.23", collection_name="foo.bar"),
     "[DEPRECATION WARNING]: hi. This feature will be removed from collection 'foo.bar' version 1.23."),
    (dict(msg="hi", date="2025-01-01", collection_name="foo.bar"),
     "[DEPRECATION WARNING]: hi. This feature will be removed from collection 'foo.bar' in a release after 2025-01-01."),
))
def test_get_deprecation_message(kw: dict[str, t.Any], expected: str) -> None:
    """Validate the deprecated public version of this function."""

    assert Display().get_deprecation_message(**kw) == expected



# -*- coding: utf-8 -*-
# (c) 2018 Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.module_utils import urls


def test_basic_auth_header():
    header = urls.basic_auth_header('user', 'passwd')
    assert header == b'Basic dXNlcjpwYXNzd2Q='


def test_ParseResultDottedDict():
    url = 'https://ansible.com/blog'
    parts = urls.urlparse(url)
    dotted_parts = urls.ParseResultDottedDict(parts._asdict())
    assert parts[0] == dotted_parts.scheme

    assert dotted_parts.as_list() == list(parts)


def test_unix_socket_patch_httpconnection_connect(mocker):
    unix_conn = mocker.patch.object(urls.UnixHTTPConnection, 'connect')
    conn = urls.http.client.HTTPConnection('ansible.com')
    with urls.unix_socket_patch_httpconnection_connect():
        conn.connect()
    assert unix_conn.call_count == 1



"""This Python module calls deprecation functions in a variety of ways to validate call inference is supported in all common scenarios."""
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
name: deprecated
short_description: lookup
description: Lookup.
author:
  - Ansible Core Team
"""

EXAMPLES = """#"""
RETURN = """#"""

import ansible.utils.display
import ansible.module_utils.common.warnings

from ansible.module_utils import datatag
from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.basic import AnsibleModule as AliasedAnsibleModule
from ansible.module_utils.basic import deprecate
from ansible.module_utils.common import warnings
from ansible.module_utils.common.warnings import deprecate as basic_deprecate
from ansible.module_utils.datatag import deprecate_value
from ansible.plugins.lookup import LookupBase
from ansible.utils import display as x_display
from ansible.utils.display import Display as XDisplay
from ansible.utils.display import _display

global_display = XDisplay()
other_global_display = x_display.Display()
foreign_global_display = x_display._display

# extra lines below to allow for adding more imports without shifting the line numbers of the code that follows
#
#
#
#
#
#


class LookupModule(LookupBase):
    def run(self, **kwargs):
        return []


class MyModule(AnsibleModule):
    """A class."""

    do_deprecated = global_display.deprecated

    def my_method(self) -> None:
        """A method."""

        self.deprecate('', version='2.0.0', collection_name='ns.col')


def give_me_a_func():
    return global_display.deprecated


def do_stuff() -> None:
    """A function."""
    d1 = x_display.Display()
    d2 = XDisplay()

    MyModule.do_deprecated('', version='2.0.0', collection_name='ns.col')
    basic_deprecate('', version='2.0.0', collection_name='ns.col')
    ansible.utils.display._display.deprecated('', version='2.0.0', collection_name='ns.col')
    d1.deprecated('', version='2.0.0', collection_name='ns.col')
    d2.deprecated('', version='2.0.0', collection_name='ns.col')
    x_display.Display().deprecated('', version='2.0.0', collection_name='ns.col')
    XDisplay().deprecated('', version='2.0.0', collection_name='ns.col')
    warnings.deprecate('', version='2.0.0', collection_name='ns.col')
    deprecate('', version='2.0.0', collection_name='ns.col')
    datatag.deprecate_value("thing", '', collection_name='ns.col', version='2.0.0')
    deprecate_value("thing", '', collection_name='ns.col', version='2.0.0')
    global_display.deprecated('', version='2.0.0', collection_name='ns.col')
    other_global_display.deprecated('', version='2.0.0', collection_name='ns.col')
    foreign_global_display.deprecated('', version='2.0.0', collection_name='ns.col')
    _display.deprecated('', version='2.0.0', collection_name='ns.col')
    give_me_a_func()("hello")  # not detected

    # don't validate date/version/collection_name if all are dynamic
    a_msg = 'message not checked'
    a_date = 'date not checked'
    a_version = 'version not checked'
    a_collection_name = 'mismatched'
    _display.deprecated(msg=a_msg, date=a_date, version=a_version, collection_name=a_collection_name)

    wrapper = MyWrapper(AnsibleModule({}))
    wrapper.module.deprecate('', version='2.0.0', collection_name='ns.col')

    wrapper = MyAliasedWrapper(AnsibleModule({}))
    wrapper.module.deprecate('', version='2.0.0', collection_name='ns.col')

    wrapper = MyOtherWrapper(AnsibleModule({}))
    wrapper.module.deprecate('', version='2.0.0', collection_name='ns.col')

    wrapper = MyOtherAliasedWrapper(AnsibleModule({}))
    wrapper.module.deprecate('', version='2.0.0', collection_name='ns.col')

    wrapper = MyTypeCommentWrapper(AnsibleModule({}))
    wrapper.module.deprecate('', version='2.0.0', collection_name='ns.col')


class MyWrapper:
    def __init__(self, thing) -> None:
        self.module: AnsibleModule = thing


class MyAliasedWrapper:
    def __init__(self, thing) -> None:
        self.module: AliasedAnsibleModule = thing


class MyOtherWrapper:
    def __init__(self, thing: AnsibleModule) -> None:
        self.module = thing


class MyOtherAliasedWrapper:
    def __init__(self, thing: AliasedAnsibleModule) -> None:
        self.module = thing


class MyTypeCommentWrapper:
    def __init__(self, thing) -> None:
        self.module = thing  # type: AnsibleModule



from __future__ import annotations

import ast
import pathlib
import sys


class CallVisitor(ast.NodeVisitor):
    def __init__(self, path: str) -> None:
        self.path = path

    def visit_Call(self, node: ast.Call) -> None:
        if isinstance(node.func, ast.Name) and node.func.id.endswith("FieldAttribute"):
            if len([kw for kw in node.keywords if kw.arg in ("default", "required")]) > 1:
                print(f"{self.path}:{node.lineno}:{node.col_offset}: use only one of `default` or `required` with `{node.func.id}`")


def main() -> None:
    for path in sys.argv[1:] or sys.stdin.read().splitlines():
        tree = ast.parse(pathlib.Path(path).read_text(), path)
        CallVisitor(path).visit(tree)


if __name__ == "__main__":
    main()



from __future__ import annotations

from ansible.plugins.action import ActionBase


class ActionModule(ActionBase):
    def run(self, tmp=None, task_vars=None):
        self._display.deprecated("a deprecation warning", version="2.99")
        self._display.warning("a warning")

        return dict(changed=False)



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest
from unittest.mock import MagicMock

from ansible.executor.playbook_executor import PlaybookExecutor
from ansible.playbook import Playbook
from ansible._internal._templating._engine import TemplateEngine
from ansible.utils import context_objects as co

from units.mock.loader import DictDataLoader


class TestPlaybookExecutor(unittest.TestCase):

    def setUp(self):
        # Reset command line args for every test
        co.GlobalCLIArgs._Singleton__instance = None

    def tearDown(self):
        # And cleanup after ourselves too
        co.GlobalCLIArgs._Singleton__instance = None

    def test_get_serialized_batches(self):
        fake_loader = DictDataLoader({
            'no_serial.yml': """
            - hosts: all
              gather_facts: no
              tasks:
              - debug: var=inventory_hostname
            """,
            'serial_int.yml': """
            - hosts: all
              gather_facts: no
              serial: 2
              tasks:
              - debug: var=inventory_hostname
            """,
            'serial_pct.yml': """
            - hosts: all
              gather_facts: no
              serial: 20%
              tasks:
              - debug: var=inventory_hostname
            """,
            'serial_list.yml': """
            - hosts: all
              gather_facts: no
              serial: [1, 2, 3]
              tasks:
              - debug: var=inventory_hostname
            """,
            'serial_list_mixed.yml': """
            - hosts: all
              gather_facts: no
              serial: [1, "20%", -1]
              tasks:
              - debug: var=inventory_hostname
            """,
        })

        mock_inventory = MagicMock()
        mock_var_manager = MagicMock()

        templar = TemplateEngine(loader=fake_loader)

        pbe = PlaybookExecutor(
            playbooks=['no_serial.yml', 'serial_int.yml', 'serial_pct.yml', 'serial_list.yml', 'serial_list_mixed.yml'],
            inventory=mock_inventory,
            variable_manager=mock_var_manager,
            loader=fake_loader,
            passwords=[],
        )

        playbook = Playbook.load(pbe._playbooks[0], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']
        self.assertEqual(pbe._get_serialized_batches(play), [['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']])

        playbook = Playbook.load(pbe._playbooks[1], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']
        self.assertEqual(
            pbe._get_serialized_batches(play),
            [['host0', 'host1'], ['host2', 'host3'], ['host4', 'host5'], ['host6', 'host7'], ['host8', 'host9']]
        )

        playbook = Playbook.load(pbe._playbooks[2], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']
        self.assertEqual(
            pbe._get_serialized_batches(play),
            [['host0', 'host1'], ['host2', 'host3'], ['host4', 'host5'], ['host6', 'host7'], ['host8', 'host9']]
        )

        playbook = Playbook.load(pbe._playbooks[3], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']
        self.assertEqual(
            pbe._get_serialized_batches(play),
            [['host0'], ['host1', 'host2'], ['host3', 'host4', 'host5'], ['host6', 'host7', 'host8'], ['host9']]
        )

        playbook = Playbook.load(pbe._playbooks[4], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']
        self.assertEqual(pbe._get_serialized_batches(play), [['host0'], ['host1', 'host2'], ['host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9']])

        # Test when serial percent is under 1.0
        playbook = Playbook.load(pbe._playbooks[2], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2']
        self.assertEqual(pbe._get_serialized_batches(play), [['host0'], ['host1'], ['host2']])

        # Test when there is a remainder for serial as a percent
        playbook = Playbook.load(pbe._playbooks[2], variable_manager=mock_var_manager, loader=fake_loader)
        play = playbook.get_plays()[0]
        play.post_validate(templar)
        mock_inventory.get_hosts.return_value = ['host0', 'host1', 'host2', 'host3', 'host4', 'host5', 'host6', 'host7', 'host8', 'host9', 'host10']
        self.assertEqual(
            pbe._get_serialized_batches(play),
            [['host0', 'host1'], ['host2', 'host3'], ['host4', 'host5'], ['host6', 'host7'], ['host8', 'host9'], ['host10']]
        )



"""Support code for CI environments."""

from __future__ import annotations

import abc
import dataclasses
import datetime
import json
import pathlib
import tempfile
import typing as t

from ..config import (
    CommonConfig,
    TestConfig,
)

from ..util import (
    ApplicationError,
    display,
    get_subclasses,
    import_plugins,
    raw_command,
    cache,
)


@dataclasses.dataclass(frozen=True, kw_only=True)
class AuthContext:
    """Information about the request to which authentication will be applied."""

    stage: str
    provider: str
    request_id: str


class AuthHelper:
    """Authentication helper."""

    NAMESPACE: t.ClassVar = 'ci@core.ansible.com'

    def __init__(self, key_file: pathlib.Path) -> None:
        self.private_key_file = pathlib.Path(str(key_file).removesuffix('.pub'))
        self.public_key_file = pathlib.Path(f'{self.private_key_file}.pub')

    def sign_request(self, request: dict[str, object], context: AuthContext) -> None:
        """Sign the given auth request using the provided context."""
        request.update(
            stage=context.stage,
            provider=context.provider,
            request_id=context.request_id,
            timestamp=datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0).isoformat(),
        )

        with tempfile.TemporaryDirectory() as temp_dir:
            payload_path = pathlib.Path(temp_dir) / 'auth.json'
            payload_path.write_text(json.dumps(request, sort_keys=True))

            cmd = ['ssh-keygen', '-q', '-Y', 'sign', '-f', str(self.private_key_file), '-n', self.NAMESPACE, str(payload_path)]
            raw_command(cmd, capture=False, interactive=True)

            signature_path = pathlib.Path(f'{payload_path}.sig')
            signature = signature_path.read_text()

        request.update(signature=signature)


class GeneratingAuthHelper(AuthHelper, metaclass=abc.ABCMeta):
    """Authentication helper which generates a key pair on demand."""

    def __init__(self) -> None:
        super().__init__(pathlib.Path('~/.ansible/test/ansible-core-ci').expanduser())

    def sign_request(self, request: dict[str, object], context: AuthContext) -> None:
        if not self.private_key_file.exists():
            self.generate_key_pair()

        super().sign_request(request, context)

    def generate_key_pair(self) -> None:
        """Generate key pair."""
        self.private_key_file.parent.mkdir(parents=True, exist_ok=True)

        raw_command(['ssh-keygen', '-q', '-f', str(self.private_key_file), '-N', ''], capture=True)


class ChangeDetectionNotSupported(ApplicationError):
    """Exception for cases where change detection is not supported."""


class CIProvider(metaclass=abc.ABCMeta):
    """Base class for CI provider plugins."""

    priority = 500

    @staticmethod
    @abc.abstractmethod
    def is_supported() -> bool:
        """Return True if this provider is supported in the current running environment."""

    @property
    @abc.abstractmethod
    def code(self) -> str:
        """Return a unique code representing this provider."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Return descriptive name for this provider."""

    @abc.abstractmethod
    def generate_resource_prefix(self) -> str:
        """Return a resource prefix specific to this CI provider."""

    @abc.abstractmethod
    def get_base_commit(self, args: CommonConfig) -> str:
        """Return the base commit or an empty string."""

    @abc.abstractmethod
    def detect_changes(self, args: TestConfig) -> t.Optional[list[str]]:
        """Initialize change detection."""

    @abc.abstractmethod
    def supports_core_ci_auth(self) -> bool:
        """Return True if Ansible Core CI is supported."""

    @abc.abstractmethod
    def prepare_core_ci_request(self, config: dict[str, object], context: AuthContext) -> dict[str, object]:
        """Prepare an Ansible Core CI request using the given config and context."""

    @abc.abstractmethod
    def get_git_details(self, args: CommonConfig) -> t.Optional[dict[str, t.Any]]:
        """Return details about git in the current environment."""


@cache
def get_ci_provider() -> CIProvider:
    """Return a CI provider instance for the current environment."""
    provider = None

    import_plugins('ci')

    candidates = sorted(get_subclasses(CIProvider), key=lambda subclass: (subclass.priority, subclass.__name__))

    for candidate in candidates:
        if candidate.is_supported():
            provider = candidate()
            break

    if provider.code:
        display.info('Detected CI provider: %s' % provider.name)

    return provider



"""Infrastructure for patching callables with alternative implementations as needed based on patch-specific test criteria."""

from __future__ import annotations

import abc
import typing as t


@t.runtime_checkable
class PatchedTarget(t.Protocol):
    """Runtime-checkable protocol that allows identification of a patched function via `isinstance`."""

    unpatched_implementation: t.Callable


class CallablePatch(abc.ABC):
    """Base class for patches that provides abstractions for validation of broken behavior, installation of patches, and validation of fixed behavior."""

    target_container: t.ClassVar
    """The module object containing the function to be patched."""

    target_attribute: t.ClassVar[str]
    """The attribute name on the target module to patch."""

    unpatched_implementation: t.ClassVar[t.Callable]
    """The unpatched implementation. Available only after the patch has been applied."""

    @classmethod
    @abc.abstractmethod
    def is_patch_needed(cls) -> bool:
        """Returns True if the patch is currently needed. Returns False if the original target does not need the patch or the patch has already been applied."""

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> t.Any:
        """Invoke the patched or original implementation, depending on whether the patch has been applied or not."""

    @classmethod
    def is_patched(cls) -> bool:
        """Returns True if the patch has been applied, otherwise returns False."""
        return isinstance(cls.get_current_implementation(), PatchedTarget)  # using a protocol lets us be more resilient to module unload weirdness

    @classmethod
    def get_current_implementation(cls) -> t.Any:
        """Get the current (possibly patched) implementation from the patch target container."""
        return getattr(cls.target_container, cls.target_attribute)

    @classmethod
    def patch(cls) -> None:
        """Idempotently apply this patch (if needed)."""
        if cls.is_patched():
            return

        cls.unpatched_implementation = cls.get_current_implementation()

        if not cls.is_patch_needed():
            return

        # __call__ requires an instance (otherwise it'll be __new__)
        setattr(cls.target_container, cls.target_attribute, cls())

        if not cls.is_patch_needed():
            return

        setattr(cls.target_container, cls.target_attribute, cls.unpatched_implementation)

        raise RuntimeError(f"Validation of '{cls.target_container.__name__}.{cls.target_attribute}' failed after patching.")



from __future__ import annotations

import inspect
import sys

from ... import basic
from . import _respawn_wrapper


def create_payload() -> str:
    """Create and return an AnsiballZ payload for respawning a module."""
    main = sys.modules['__main__']
    code = inspect.getsource(_respawn_wrapper)

    args = dict(
        module_fqn=main._module_fqn,
        modlib_path=main._modlib_path,
        profile=basic._ANSIBLE_PROFILE,
        json_params=basic._ANSIBLE_ARGS,
    )

    args_string = '\n'.join(f'{key}={value!r},' for key, value in args.items())

    wrapper = f"""{code}

if __name__ == "__main__":
    _respawn_main(
{args_string}
)
"""

    return wrapper



"""Support fixtures for testing Ansible modules."""

from __future__ import annotations

from ansible.module_utils._internal import _traceback
from ansible.module_utils.common import warnings

import pytest


class ModuleEnvMocker:
    """
    A Swiss Army Knife mock fixture for target-side environments (e.g., modules, module_utils).

    Using this fixture automatically applies a number of default patches, and offers tests a one-stop
    fixture to configure a number of others.

    Default behaviors:
    * Warnings and deprecations are cleared.
    * Traceback collection is disabled.
    """

    def __init__(self, mp: pytest.MonkeyPatch):
        self._mp = mp
        self.reset_warnings()
        self.reset_deprecations()
        self.set_traceback_config()

    def set_traceback_config(self, value: list[_traceback.TracebackEvent] | None = None) -> None:
        """Accepts a list of TracebackEvent values to enable."""
        self._mp.setattr(_traceback, '_module_tracebacks_enabled_events', value or [])

    def reset_warnings(self) -> None:
        """Clears accumulated target-side warnings."""
        self._mp.setattr(warnings, '_global_warnings', {})

    def reset_deprecations(self) -> None:
        """Clears accumulated target-side deprecation warnings."""
        self._mp.setattr(warnings, '_global_deprecations', {})


@pytest.fixture
def module_env_mocker(monkeypatch: pytest.MonkeyPatch) -> ModuleEnvMocker:
    """Provides a ModuleEnvMocker instance that allows configurable simulation of an Ansible Python module runtime environment."""
    return ModuleEnvMocker(monkeypatch)



# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import os
import typing as t

from ansible.module_utils.facts.collector import BaseFactCollector


def is_chroot(module=None):

    is_chroot = None

    if os.environ.get('debian_chroot', False):
        is_chroot = True
    else:
        my_root = os.stat('/')
        try:
            # check if my file system is the root one
            proc_root = os.stat('/proc/1/root/.')
            is_chroot = my_root.st_ino != proc_root.st_ino or my_root.st_dev != proc_root.st_dev
        except Exception:
            # I'm not root or no proc, fallback to checking it is inode #2
            fs_root_ino = 2

            if module is not None:
                stat_path = module.get_bin_path('stat')
                if stat_path:
                    cmd = [stat_path, '-f', '--format=%T', '/']
                    rc, out, err = module.run_command(cmd)
                    if 'btrfs' in out:
                        fs_root_ino = 256
                    elif 'xfs' in out:
                        fs_root_ino = 128

            is_chroot = (my_root.st_ino != fs_root_ino)

    return is_chroot


class ChrootFactCollector(BaseFactCollector):
    name = 'chroot'
    _fact_ids = set(['is_chroot'])  # type: t.Set[str]

    def collect(self, module=None, collected_facts=None):
        return {'is_chroot': is_chroot(module)}



# -*- coding: utf-8 -*-

# Copyright: (c) 2012, Jeroen Hoekx <jeroen@hoekx.be>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations


DOCUMENTATION = r"""
---
module: wait_for
short_description: Waits for a condition before continuing
description:
     - You can wait for a set amount of time O(timeout), this is the default if nothing is specified or just O(timeout) is specified.
       This does not produce an error.
     - Waiting for a port to become available is useful for when services are not immediately available after their init scripts return
       which is true of certain Java application servers.
     - It is also useful when starting guests with the M(community.libvirt.virt) module and needing to pause until they are ready.
     - This module can also be used to wait for a regex match a string to be present in a file.
     - In Ansible 1.6 and later, this module can also be used to wait for a file to be available or
       absent on the filesystem.
     - In Ansible 1.8 and later, this module can also be used to wait for active connections to be closed before continuing, useful if a node
       is being rotated out of a load balancer pool.
     - For Windows targets, use the M(ansible.windows.win_wait_for) module instead.
version_added: "0.7"
options:
  host:
    description:
      - A resolvable hostname or IP address to wait for.
    type: str
    default: 127.0.0.1
  timeout:
    description:
      - Maximum number of seconds to wait for, when used with another condition it will force an error.
      - When used without other conditions it is equivalent of just sleeping.
    type: int
    default: 300
  connect_timeout:
    description:
      - Maximum number of seconds to wait for a connection to happen before closing and retrying.
    type: int
    default: 5
  delay:
    description:
      - Number of seconds to wait before starting to poll.
    type: int
    default: 0
  port:
    description:
      - Port number to poll.
      - O(path) and O(port) are mutually exclusive parameters.
    type: int
  active_connection_states:
    description:
      - The list of TCP connection states which are counted as active connections.
    type: list
    elements: str
    default: [ ESTABLISHED, FIN_WAIT1, FIN_WAIT2, SYN_RECV, SYN_SENT, TIME_WAIT ]
    version_added: "2.3"
  state:
    description:
      - Either V(present), V(started), or V(stopped), V(absent), or V(drained).
      - When checking a port V(started) will ensure the port is open, V(stopped) will check that it is closed, V(drained) will check for active connections.
      - When checking for a file or a search string V(present) or V(started) will ensure that the file or string is present before continuing,
        V(absent) will check that file is absent or removed.
    type: str
    choices: [ absent, drained, present, started, stopped ]
    default: started
  path:
    description:
      - Path to a file on the filesystem that must exist before continuing.
      - O(path) and O(port) are mutually exclusive parameters.
    type: path
    version_added: "1.4"
  search_regex:
    description:
      - Can be used to match a string in either a file or a socket connection.
      - Defaults to a multiline regex.
      - When inspecting a system log file and a static string, remember that Ansible by default logs its own actions there;
        see the notes and examples for information.
    type: str
    version_added: "1.4"
  exclude_hosts:
    description:
      - List of hosts or IPs to ignore when looking for active TCP connections for V(drained) state.
    type: list
    elements: str
    version_added: "1.8"
  sleep:
    description:
      - Number of seconds to sleep between checks.
      - Before Ansible 2.3 this was hardcoded to 1 second.
    type: int
    default: 1
    version_added: "2.3"
  msg:
    description:
      - This overrides the normal error message from a failure to meet the required conditions.
    type: str
    version_added: "2.4"
extends_documentation_fragment: action_common_attributes
attributes:
    check_mode:
        support: none
    diff_mode:
        support: none
    platform:
        platforms: posix
notes:
  - Under some circumstances when using mandatory access control, a path may always be treated as being absent even if it exists, but
    can't be modified or created by the remote user either.
  - When waiting for a path, symbolic links will be followed.  Many other modules that manipulate files do not follow symbolic links,
    so operations on the path using other modules may not work exactly as expected.
  - When searching a static string within a system log file, it is important to account for potential self-matching against log entries
    generated by the Ansible modules.  To prevent this, add a regular expression construct into the search string. For example, to match
    a literal string 'this thing', one could use a regular expression like 'this t[h]ing'.
seealso:
- module: ansible.builtin.wait_for_connection
- module: ansible.windows.win_wait_for
- module: community.windows.win_wait_for_process
author:
    - Jeroen Hoekx (@jhoekx)
    - John Jarvis (@jarv)
    - Andrii Radyk (@AnderEnder)
"""

EXAMPLES = r"""
- name: Sleep for 300 seconds and continue with play
  ansible.builtin.wait_for:
    timeout: 300
  delegate_to: localhost

- name: Wait for port 8000 to become open on the host, don't start checking for 10 seconds
  ansible.builtin.wait_for:
    port: 8000
    delay: 10

- name: Waits for port 8000 of any IP to close active connections, don't start checking for 10 seconds
  ansible.builtin.wait_for:
    host: 0.0.0.0
    port: 8000
    delay: 10
    state: drained

- name: Wait for port 8000 of any IP to close active connections, ignoring connections for specified hosts
  ansible.builtin.wait_for:
    host: 0.0.0.0
    port: 8000
    state: drained
    exclude_hosts: 10.2.1.2,10.2.1.3

- name: Wait until the file /tmp/foo is present before continuing
  ansible.builtin.wait_for:
    path: /tmp/foo

- name: Wait until the string "completed" is in the file /tmp/foo before continuing
  ansible.builtin.wait_for:
    path: /tmp/foo
    search_regex: completed

- name: Wait until the string "tomcat up" is in syslog, use regex character set to avoid self match
  ansible.builtin.wait_for:
    path: /var/log/syslog
    search_regex: 'tomcat [u]p'

- name: Wait until regex pattern matches in the file /tmp/foo and print the matched group
  ansible.builtin.wait_for:
    path: /tmp/foo
    search_regex: completed (?P<task>\w+)
  register: waitfor
- ansible.builtin.debug:
    msg: Completed {{ waitfor['match_groupdict']['task'] }}

- name: Wait until the lock file is removed
  ansible.builtin.wait_for:
    path: /var/lock/file.lock
    state: absent

- name: Wait until the process is finished and pid was destroyed
  ansible.builtin.wait_for:
    path: /proc/3466/status
    state: absent

- name: Output customized message when failed
  ansible.builtin.wait_for:
    path: /tmp/foo
    state: present
    msg: Timeout to find file /tmp/foo

# Do not assume the inventory_hostname is resolvable and delay 10 seconds at start
- name: Wait 300 seconds for port 22 to become open and contain "OpenSSH"
  ansible.builtin.wait_for:
    port: 22
    host: '{{ (ansible_ssh_host|default(ansible_host))|default(inventory_hostname) }}'
    search_regex: OpenSSH
    delay: 10
    timeout: 300
  delegate_to: localhost

# Same as above but using config lookup for the target,
# most plugins use 'remote_addr', but ssh uses 'host'
- name: Wait 300 seconds for port 22 to become open and contain "OpenSSH"
  ansible.builtin.wait_for:
    port: 22
    host: "{{ lookup('config', 'host', plugin_name='ssh', plugin_type='connection') }}"
    search_regex: OpenSSH
    delay: 10
    timeout: 300
  delegate_to: localhost
"""

RETURN = r"""
elapsed:
  description: The number of seconds that elapsed while waiting
  returned: always
  type: int
  sample: 23
match_groups:
  description: Tuple containing all the subgroups of the match as returned by U(https://docs.python.org/3/library/re.html#re.MatchObject.groups)
  returned: always
  type: list
  sample: ['match 1', 'match 2']
match_groupdict:
  description: Dictionary containing all the named subgroups of the match, keyed by the subgroup name,
    as returned by U(https://docs.python.org/3/library/re.html#re.MatchObject.groupdict)
  returned: always
  type: dict
  sample:
    {
      'group': 'match'
    }
"""

import binascii
import contextlib
import errno
import math
import mmap
import os
import re
import select
import socket
import time

from datetime import datetime, timedelta, timezone

from ansible.module_utils.basic import AnsibleModule, missing_required_lib
from ansible.module_utils.common.sys_info import get_platform_subclass
from ansible.module_utils.common.text.converters import to_bytes, to_native


HAS_PSUTIL = False
PSUTIL_IMP_ERR = None
try:
    import psutil
    HAS_PSUTIL = True
    # just because we can import it on Linux doesn't mean we will use it
except ImportError as ex:
    PSUTIL_IMP_ERR = ex


class TCPConnectionInfo(object):
    """
    This is a generic TCP Connection Info strategy class that relies
    on the psutil module, which is not ideal for targets, but necessary
    for cross platform support.

    A subclass may wish to override some or all of these methods.
      - _get_exclude_ips()
      - get_active_connections()

    All subclasses MUST define platform and distribution (which may be None).
    """
    platform = 'Generic'
    distribution = None

    match_all_ips = {
        socket.AF_INET: '0.0.0.0',
        socket.AF_INET6: '::',
    }
    ipv4_mapped_ipv6_address = {
        'prefix': '::ffff',
        'match_all': '::ffff:0.0.0.0'
    }

    def __new__(cls, *args, **kwargs):
        new_cls = get_platform_subclass(TCPConnectionInfo)
        return super(cls, new_cls).__new__(new_cls)

    def __init__(self, module):
        self.module = module
        self.ips = _convert_host_to_ip(module.params['host'])
        self.port = int(self.module.params['port'])
        self.exclude_ips = self._get_exclude_ips()
        if not HAS_PSUTIL:
            module.fail_json(msg=missing_required_lib('psutil'), exception=PSUTIL_IMP_ERR)

    def _get_exclude_ips(self):
        exclude_hosts = self.module.params['exclude_hosts']
        exclude_ips = []
        if exclude_hosts is not None:
            for host in exclude_hosts:
                exclude_ips.extend(_convert_host_to_ip(host))
        return exclude_ips

    def get_active_connections_count(self):
        active_connections = 0
        for p in psutil.process_iter():
            try:
                if hasattr(p, 'get_connections'):
                    connections = p.get_connections(kind='inet')
                else:
                    connections = p.connections(kind='inet')
            except psutil.Error:
                # Process is Zombie or other error state
                continue
            for conn in connections:
                if conn.status not in self.module.params['active_connection_states']:
                    continue
                if hasattr(conn, 'local_address'):
                    (local_ip, local_port) = conn.local_address
                else:
                    (local_ip, local_port) = conn.laddr
                if self.port != local_port:
                    continue
                if hasattr(conn, 'remote_address'):
                    (remote_ip, remote_port) = conn.remote_address
                else:
                    (remote_ip, remote_port) = conn.raddr
                if (conn.family, remote_ip) in self.exclude_ips:
                    continue
                if any((
                    (conn.family, local_ip) in self.ips,
                    (conn.family, self.match_all_ips[conn.family]) in self.ips,
                    local_ip.startswith(self.ipv4_mapped_ipv6_address['prefix']) and
                        (conn.family, self.ipv4_mapped_ipv6_address['match_all']) in self.ips,
                )):
                    active_connections += 1
        return active_connections


# ===========================================
# Subclass: Linux

class LinuxTCPConnectionInfo(TCPConnectionInfo):
    """
    This is a TCP Connection Info evaluation strategy class
    that utilizes information from Linux's procfs. While less universal,
    does allow Linux targets to not require an additional library.
    """
    platform = 'Linux'
    distribution = None

    source_file = {
        socket.AF_INET: '/proc/net/tcp',
        socket.AF_INET6: '/proc/net/tcp6'
    }
    match_all_ips = {
        socket.AF_INET: '00000000',
        socket.AF_INET6: '00000000000000000000000000000000',
    }
    ipv4_mapped_ipv6_address = {
        'prefix': '0000000000000000FFFF0000',
        'match_all': '0000000000000000FFFF000000000000'
    }
    local_address_field = 1
    remote_address_field = 2
    connection_state_field = 3

    def __init__(self, module):
        self.module = module
        self.ips = _convert_host_to_hex(module.params['host'])
        self.port = "%0.4X" % int(module.params['port'])
        self.exclude_ips = self._get_exclude_ips()

    def _get_exclude_ips(self):
        exclude_hosts = self.module.params['exclude_hosts']
        exclude_ips = []
        if exclude_hosts is not None:
            for host in exclude_hosts:
                exclude_ips.extend(_convert_host_to_hex(host))
        return exclude_ips

    def get_active_connections_count(self):
        active_connections = 0
        for family in self.source_file.keys():
            if not os.path.isfile(self.source_file[family]):
                continue
            try:
                with open(self.source_file[family]) as f:
                    for tcp_connection in f.readlines():
                        tcp_connection = tcp_connection.strip().split()
                        if tcp_connection[self.local_address_field] == 'local_address':
                            continue
                        if (tcp_connection[self.connection_state_field] not in
                                [get_connection_state_id(_connection_state) for _connection_state in self.module.params['active_connection_states']]):
                            continue
                        (local_ip, local_port) = tcp_connection[self.local_address_field].split(':')
                        if self.port != local_port:
                            continue
                        (remote_ip, remote_port) = tcp_connection[self.remote_address_field].split(':')
                        if (family, remote_ip) in self.exclude_ips:
                            continue
                        if any((
                            (family, local_ip) in self.ips,
                            (family, self.match_all_ips[family]) in self.ips,
                            local_ip.startswith(self.ipv4_mapped_ipv6_address['prefix']) and
                                (family, self.ipv4_mapped_ipv6_address['match_all']) in self.ips,
                        )):
                            active_connections += 1
            except OSError:
                pass

        return active_connections


def _convert_host_to_ip(host):
    """
    Perform forward DNS resolution on host, IP will give the same IP

    Args:
        host: String with either hostname, IPv4, or IPv6 address

    Returns:
        List of tuples containing address family and IP
    """
    addrinfo = socket.getaddrinfo(host, 80, 0, 0, socket.SOL_TCP)
    ips = []
    for family, socktype, proto, canonname, sockaddr in addrinfo:
        ip = sockaddr[0]
        ips.append((family, ip))
        if family == socket.AF_INET:
            ips.append((socket.AF_INET6, "::ffff:" + ip))
    return ips


def _convert_host_to_hex(host):
    """
    Convert the provided host to the format in /proc/net/tcp*

    /proc/net/tcp uses little-endian four byte hex for ipv4
    /proc/net/tcp6 uses little-endian per 4B word for ipv6

    Args:
        host: String with either hostname, IPv4, or IPv6 address

    Returns:
        List of tuples containing address family and the
        little-endian converted host
    """
    ips = []
    if host is not None:
        for family, ip in _convert_host_to_ip(host):
            hexip_nf = binascii.b2a_hex(socket.inet_pton(family, ip))
            hexip_hf = ""
            for i in range(0, len(hexip_nf), 8):
                ipgroup_nf = hexip_nf[i:i + 8]
                ipgroup_hf = socket.ntohl(int(ipgroup_nf, base=16))
                hexip_hf = "%s%08X" % (hexip_hf, ipgroup_hf)
            ips.append((family, hexip_hf))
    return ips


def _timedelta_total_seconds(timedelta):
    return (
        timedelta.microseconds + 0.0 +
        (timedelta.seconds + timedelta.days * 24 * 3600) * 10 ** 6) / 10 ** 6


def get_connection_state_id(state):
    connection_state_id = {
        'ESTABLISHED': '01',
        'SYN_SENT': '02',
        'SYN_RECV': '03',
        'FIN_WAIT1': '04',
        'FIN_WAIT2': '05',
        'TIME_WAIT': '06',
    }
    return connection_state_id[state]


def main():

    module = AnsibleModule(
        argument_spec=dict(
            host=dict(type='str', default='127.0.0.1'),
            timeout=dict(type='int', default=300),
            connect_timeout=dict(type='int', default=5),
            delay=dict(type='int', default=0),
            port=dict(type='int'),
            active_connection_states=dict(type='list', elements='str', default=['ESTABLISHED', 'FIN_WAIT1', 'FIN_WAIT2', 'SYN_RECV', 'SYN_SENT', 'TIME_WAIT']),
            path=dict(type='path'),
            search_regex=dict(type='str'),
            state=dict(type='str', default='started', choices=['absent', 'drained', 'present', 'started', 'stopped']),
            exclude_hosts=dict(type='list', elements='str'),
            sleep=dict(type='int', default=1),
            msg=dict(type='str'),
        ),
    )

    host = module.params['host']
    timeout = module.params['timeout']
    connect_timeout = module.params['connect_timeout']
    delay = module.params['delay']
    port = module.params['port']
    state = module.params['state']

    path = module.params['path']
    b_path = to_bytes(path, errors='surrogate_or_strict', nonstring='passthru')

    search_regex = module.params['search_regex']
    b_search_regex = to_bytes(search_regex, errors='surrogate_or_strict', nonstring='passthru')

    msg = module.params['msg']

    if search_regex is not None:
        try:
            b_compiled_search_re = re.compile(b_search_regex, re.MULTILINE)
        except re.error as e:
            module.fail_json(msg="Invalid regular expression: %s" % e)
    else:
        b_compiled_search_re = None

    match_groupdict = {}
    match_groups = ()

    if port and path:
        module.fail_json(msg="port and path parameter can not both be passed to wait_for", elapsed=0)
    if path and state == 'stopped':
        module.fail_json(msg="state=stopped should only be used for checking a port in the wait_for module", elapsed=0)
    if path and state == 'drained':
        module.fail_json(msg="state=drained should only be used for checking a port in the wait_for module", elapsed=0)
    if module.params['exclude_hosts'] is not None and state != 'drained':
        module.fail_json(msg="exclude_hosts should only be with state=drained", elapsed=0)
    for _connection_state in module.params['active_connection_states']:
        try:
            get_connection_state_id(_connection_state)
        except Exception:
            module.fail_json(msg="unknown active_connection_state (%s) defined" % _connection_state, elapsed=0)

    start = datetime.now(timezone.utc)

    if delay:
        time.sleep(delay)

    if not port and not path and state != 'drained':
        time.sleep(timeout)
    elif state in ['absent', 'stopped']:
        # first wait for the stop condition
        end = start + timedelta(seconds=timeout)

        while datetime.now(timezone.utc) < end:
            if path:
                try:
                    if not os.access(b_path, os.F_OK):
                        break
                except OSError:
                    break
            elif port:
                try:
                    s = socket.create_connection((host, port), connect_timeout)
                    s.shutdown(socket.SHUT_RDWR)
                    s.close()
                except Exception:
                    break
            # Conditions not yet met, wait and try again
            time.sleep(module.params['sleep'])
        else:
            elapsed = datetime.now(timezone.utc) - start
            if port:
                module.fail_json(msg=msg or "Timeout when waiting for %s:%s to stop." % (host, port), elapsed=elapsed.seconds)
            elif path:
                module.fail_json(msg=msg or "Timeout when waiting for %s to be absent." % (path), elapsed=elapsed.seconds)

    elif state in ['started', 'present']:
        # wait for start condition
        end = start + timedelta(seconds=timeout)
        while datetime.now(timezone.utc) < end:
            if path:
                try:
                    os.stat(b_path)
                except OSError as e:
                    # If anything except file not present, throw an error
                    if e.errno != 2:
                        elapsed = datetime.now(timezone.utc) - start
                        module.fail_json(msg=msg or "Failed to stat %s, %s" % (path, e.strerror), elapsed=elapsed.seconds)
                    # file doesn't exist yet, so continue
                else:
                    # File exists.  Are there additional things to check?
                    if not b_compiled_search_re:
                        # nope, succeed!
                        break

                    try:
                        with open(b_path, 'rb') as f:
                            try:
                                with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as mm:
                                    search = b_compiled_search_re.search(mm)
                                    if search:
                                        if search.groupdict():
                                            match_groupdict = search.groupdict()
                                        if search.groups():
                                            match_groups = search.groups()
                                        break
                            except (ValueError, OSError) as e:
                                module.debug('wait_for failed to use mmap on "%s": %s. Falling back to file read().' % (path, to_native(e)))
                                # cannot mmap this file, try normal read
                                search = re.search(b_compiled_search_re, f.read())
                                if search:
                                    if search.groupdict():
                                        match_groupdict = search.groupdict()
                                    if search.groups():
                                        match_groups = search.groups()
                                    break
                            except Exception as e:
                                module.warn('wait_for failed on "%s", unexpected exception(%s): %s.).' % (path, to_native(e.__class__), to_native(e)))
                    except OSError:
                        pass
            elif port:
                alt_connect_timeout = math.ceil(
                    _timedelta_total_seconds(end - datetime.now(timezone.utc)),
                )
                try:
                    s = socket.create_connection((host, int(port)), min(connect_timeout, alt_connect_timeout))
                except Exception:
                    # Failed to connect by connect_timeout. wait and try again
                    pass
                else:
                    # Connected -- are there additional conditions?
                    if b_compiled_search_re:
                        b_data = b''
                        matched = False
                        while datetime.now(timezone.utc) < end:
                            max_timeout = math.ceil(
                                _timedelta_total_seconds(
                                    end - datetime.now(timezone.utc),
                                ),
                            )
                            readable = select.select([s], [], [], max_timeout)[0]
                            if not readable:
                                # No new data.  Probably means our timeout
                                # expired
                                continue
                            response = s.recv(1024)
                            if not response:
                                # Server shutdown
                                break
                            b_data += response
                            if b_compiled_search_re.search(b_data):
                                matched = True
                                break

                        # Shutdown the client socket
                        try:
                            s.shutdown(socket.SHUT_RDWR)
                        except OSError as ex:
                            if ex.errno != errno.ENOTCONN:
                                raise
                        # else, the server broke the connection on its end, assume it's not ready
                        else:
                            s.close()
                        if matched:
                            # Found our string, success!
                            break
                    else:
                        # Connection established, success!
                        try:
                            s.shutdown(socket.SHUT_RDWR)
                        except OSError as ex:
                            if ex.errno != errno.ENOTCONN:
                                raise
                        # else, the server broke the connection on its end, assume it's not ready
                        else:
                            s.close()
                        break

            # Conditions not yet met, wait and try again
            time.sleep(module.params['sleep'])

        else:   # while-else
            # Timeout expired
            elapsed = datetime.now(timezone.utc) - start
            if port:
                if search_regex:
                    module.fail_json(msg=msg or "Timeout when waiting for search string %s in %s:%s" % (search_regex, host, port), elapsed=elapsed.seconds)
                else:
                    module.fail_json(msg=msg or "Timeout when waiting for %s:%s" % (host, port), elapsed=elapsed.seconds)
            elif path:
                if search_regex:
                    module.fail_json(msg=msg or "Timeout when waiting for search string %s in %s" % (search_regex, path), elapsed=elapsed.seconds)
                else:
                    module.fail_json(msg=msg or "Timeout when waiting for file %s" % (path), elapsed=elapsed.seconds)

    elif state == 'drained':
        # wait until all active connections are gone
        end = start + timedelta(seconds=timeout)
        tcpconns = TCPConnectionInfo(module)
        while datetime.now(timezone.utc) < end:
            if tcpconns.get_active_connections_count() == 0:
                break

            # Conditions not yet met, wait and try again
            time.sleep(module.params['sleep'])
        else:
            elapsed = datetime.now(timezone.utc) - start
            module.fail_json(msg=msg or "Timeout when waiting for %s:%s to drain" % (host, port), elapsed=elapsed.seconds)

    elapsed = datetime.now(timezone.utc) - start
    module.exit_json(state=state, port=port, search_regex=search_regex, match_groups=match_groups, match_groupdict=match_groupdict, path=path,
                     elapsed=elapsed.seconds)


if __name__ == '__main__':
    main()



from __future__ import annotations

# exercise relative imports in package init; they behave differently
from .mod_in_subpkg_with_init import thingtocall as submod_thingtocall
from ..subpkg.submod import thingtocall as cousin_submod_thingtocall  # pylint: disable=relative-beyond-top-level


def thingtocall():
    return "thingtocall in subpkg_with_init"



from __future__ import annotations

import typing as t


class FilterModule:
    def filters(self) -> dict[str, t.Callable]:
        raise NotImplementedError()  # this should generate a warning, but shouldn't prevent other plugins from working



from __future__ import annotations

import http.server
import socketserver
import sys

if __name__ == '__main__':
    PORT = int(sys.argv[1])

    class Handler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            if self.path == '/incompleteRead':
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.send_header("Content-Length", "100")
                self.end_headers()
                self.wfile.write(b"ABCD")
            else:
                super().do_GET()

    Handler.extensions_map['.json'] = 'application/json'
    httpd = socketserver.TCPServer(("", PORT), Handler)
    httpd.serve_forever()



from __future__ import annotations

from ansible.module_utils.datatag import deprecate_value, deprecator_from_collection_name


def get_deprecation_kwargs() -> list[dict[str, object]]:
    return [
        dict(msg="Deprecation that passes collection_name, version, and help_text.", version='9999.9', collection_name='bla.bla', help_text="Help text."),
        dict(
            msg="Deprecation that passes deprecator and datetime.date.",
            date='2034-01-02',
            deprecator=deprecator_from_collection_name('bla.bla'),
        ),
        dict(msg="Deprecation that passes deprecator and string date.", date='2034-01-02', deprecator=deprecator_from_collection_name('bla.bla')),
        dict(msg="Deprecation that passes no deprecator, collection name, or date/version."),
    ]


def get_deprecated_value() -> str:
    return deprecate_value(  # pylint: disable=ansible-deprecated-unnecessary-collection-name,ansible-deprecated-collection-name-not-permitted
        value='a deprecated value',
        msg="value is deprecated",
        collection_name='foo.bar',
        version='9999.9',
    )



from __future__ import annotations

import contextlib
import signal
import types
import typing as _t

from ansible.module_utils import datatag


class AnsibleTimeoutError(BaseException):
    """A general purpose timeout."""

    _MAX_TIMEOUT = 100_000_000
    """
    The maximum supported timeout value.
    This value comes from BSD's alarm limit, which is due to that function using setitimer.
    """

    def __init__(self, timeout: int) -> None:
        self.timeout = timeout

        super().__init__(f"Timed out after {timeout} second(s).")

    @classmethod
    @contextlib.contextmanager
    def alarm_timeout(cls, timeout: int | None) -> _t.Iterator[None]:
        """
        Context for running code under an optional timeout.
        Raises an instance of this class if the timeout occurs.

        New usages of this timeout mechanism are discouraged.
        """
        if timeout is not None:
            if not isinstance(timeout, int):
                raise TypeError(f"Timeout requires 'int' argument, not {datatag.native_type_name(timeout)!r}.")

            if timeout < 0 or timeout > cls._MAX_TIMEOUT:
                # On BSD based systems, alarm is implemented using setitimer.
                # If out-of-bounds values are passed to alarm, they will return -1, which would be interpreted as an existing timer being set.
                # To avoid that, bounds checking is performed in advance.
                raise ValueError(f'Timeout {timeout} is invalid, it must be between 0 and {cls._MAX_TIMEOUT}.')

        if not timeout:
            yield  # execute the context manager's body
            return  # no timeout to deal with, exit immediately

        def on_alarm(_signal: int, _frame: types.FrameType) -> None:
            raise cls(timeout)

        if signal.signal(signal.SIGALRM, on_alarm):
            raise RuntimeError("An existing alarm handler was present.")

        try:
            try:
                if signal.alarm(timeout):
                    raise RuntimeError("An existing alarm was set.")

                yield  # execute the context manager's body
            finally:
                # Disable the alarm.
                # If the alarm fires inside this finally block, the alarm is still disabled.
                # This guarantees the cleanup code in the outer finally block runs without risk of encountering the `TaskTimeoutError` from the alarm.
                signal.alarm(0)
        finally:
            signal.signal(signal.SIGALRM, signal.SIG_DFL)



from __future__ import annotations

import importlib
import importlib.util
import types

import typing as t

from ansible.module_utils._internal._json._profiles import AnsibleProfileJSONEncoder, AnsibleProfileJSONDecoder, _JSONSerializationProfile
from ansible.module_utils import _internal

_T = t.TypeVar('_T', AnsibleProfileJSONEncoder, AnsibleProfileJSONDecoder)


def get_encoder_decoder(profile: str | types.ModuleType, return_type: type[_T]) -> type[_T]:
    class_name = 'Encoder' if return_type is AnsibleProfileJSONEncoder else 'Decoder'

    return getattr(get_serialization_module(profile), class_name)


def get_module_serialization_profile_name(name: str, controller_to_module: bool) -> str:
    if controller_to_module:
        name = f'module_{name}_c2m'
    else:
        name = f'module_{name}_m2c'

    return name


def get_module_serialization_profile_module_name(name: str, controller_to_module: bool) -> str:
    return get_serialization_module_name(get_module_serialization_profile_name(name, controller_to_module))


def get_serialization_profile(name: str | types.ModuleType) -> _JSONSerializationProfile:
    return getattr(get_serialization_module(name), '_Profile')


def get_serialization_module(name: str | types.ModuleType) -> types.ModuleType:
    return importlib.import_module(get_serialization_module_name(name))


def get_serialization_module_name(name: str | types.ModuleType) -> str:
    if isinstance(name, str):
        if '.' in name:
            return name  # name is already fully qualified

        target_name = f'{__name__}._profiles._{name}'
    elif isinstance(name, types.ModuleType):
        return name.__name__
    else:
        raise TypeError(f'Name is {type(name)} instead of {str} or {types.ModuleType}.')

    if importlib.util.find_spec(target_name):
        return target_name

    # the value of is_controller can change after import; always pick it up from the module
    if _internal.is_controller:
        controller_name = f'ansible._internal._json._profiles._{name}'

        if importlib.util.find_spec(controller_name):
            return controller_name

    raise ValueError(f'Unknown profile name {name!r}.')



# Try to globally patch Templar trust check failures to be fatal for all unit tests

from __future__ import annotations

import pytest
import sys
import typing as t

import pytest_mock

try:
    from ansible import _internal  # sets is_controller=True in controller context
    from ansible.module_utils._internal import is_controller  # allow checking is_controller
    from ansible._internal._templating._jinja_common import _TemplateConfig
    from ansible._internal._errors._handler import ErrorHandler, ErrorAction
except ImportError:
    # likely doing only module_utils testing; ignore here and rely on test_templar::test_trust_fail_raises_in_tests to ensure the right behavior
    pass
else:
    assert _internal
    assert is_controller

    # Ensure unit tests fail when encountering untrusted templates to reduce mistakes in tests.
    # Tests that need to ignore or warn on untrusted templates will need to override this setting.
    _TemplateConfig.untrusted_template_handler = ErrorHandler(ErrorAction.ERROR)

    from .controller_only_conftest import *  # pylint: disable=wildcard-import,unused-wildcard-import

from ansible.module_utils import _internal as _module_utils_internal
from ansible.module_utils._internal import _traceback as _module_utils_internal_traceback


def pytest_configure(config: pytest.Config):
    config.addinivalue_line("markers", "autoparam(value): metadata-driven parametrization")
    config.addinivalue_line("markers", "allow_delazify: test will delazify the result")


@pytest.fixture
def collection_loader() -> t.Iterator[None]:
    """
    Provide a collection loader with no collections.
    Useful for tests that fail without a collection loader, but that don't actually depend on collections.
    """
    from ansible.utils.collection_loader._collection_finder import _AnsibleCollectionFinder

    _AnsibleCollectionFinder()._install()

    try:
        yield
    finally:
        _AnsibleCollectionFinder._remove()

        for module_to_nuke in [m for m in sys.modules if m.startswith('ansible_collections')]:
            sys.modules.pop(module_to_nuke)


# @pytest.fixture(autouse=True)
# def prevent_collection_loader_leak(request: pytest.FixtureRequest):
#     # DTFIX-FUTURE: enable this fixture to ensure the collection loader has not "leaked"
#     for finder in sys.meta_path:
#         if "_AnsibleCollectionFinder" in type(finder).__name__:
#             finder._remove()
#             assert False, f"a finder was active before test {request.node.name}"
#
#     yield
#
#     for finder in sys.meta_path:
#         if "_AnsibleCollectionFinder" in type(finder).__name__:
#             finder._remove()
#             assert False, f"a finder was active after test {request.node.name}"


def pytest_collection_finish(session: pytest.Session):
    """
    This hook ensures that a collection loader is not installed after test import/collection.
    The presence of a collection loader pollutes test state in various undesirable ways.
    """
    for finder in sys.meta_path:
        if "_AnsibleCollectionFinder" in type(finder).__name__:
            assert False, "a collection loader was active after collection"


@pytest.fixture
def as_target(mocker: pytest_mock.MockerFixture) -> None:
    """Force execution in the context of a target host instead of the controller."""
    mocker.patch.object(_module_utils_internal, 'is_controller', False)
    mocker.patch.object(_module_utils_internal_traceback, '_is_traceback_enabled', _module_utils_internal_traceback._is_module_traceback_enabled)



# coding: utf-8
# (c) 2015, Toshio Kuratomi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import collections.abc as c
import datetime
import typing as t

import pytest
import yaml

from collections.abc import Sequence, Set, Mapping
from io import StringIO

import unittest

from ansible import errors
from ansible.parsing.utils.yaml import from_yaml
from ansible.parsing.yaml.dumper import AnsibleDumper
from ansible.parsing.yaml.loader import AnsibleLoader
from ansible._internal._yaml._errors import AnsibleYAMLParserError
from ansible.parsing import vault
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate
from ansible.module_utils._internal._datatag import _untaggable_types

from units.mock.yaml_helper import YamlTestUtils
from units.mock.vault_helper import TextVaultSecret

from yaml.parser import ParserError
from yaml.scanner import ScannerError

file_name = '/some/test/path/myfile.yml'


class TestAnsibleLoaderBasic(unittest.TestCase):

    def test_parse_number(self):
        stream = StringIO(u"""
                1
                """)
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, 1)
        # No line/column info saved yet

    def test_parse_string(self):
        stream = StringIO(u"""
                Ansible
                """)
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, u'Ansible')
        self.assertIsInstance(data, str)

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=2, col_num=17))

    def test_parse_utf8_string(self):
        stream = StringIO(u"""
                Cafè Eñyei
                """)
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, u'Cafè Eñyei')
        self.assertIsInstance(data, str)

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=2, col_num=17))

    def test_parse_dict(self):
        stream = StringIO(u"""
                webster: daniel
                oed: oxford
                """)
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, {'webster': 'daniel', 'oed': 'oxford'})
        self.assertEqual(len(data), 2)
        self.assertIsInstance(list(data.keys())[0], str)
        self.assertIsInstance(list(data.values())[0], str)

        # Beginning of the first key
        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=2, col_num=17))

        self.assertEqual(Origin.get_tag(data[u'webster']), Origin(path=file_name, line_num=2, col_num=26))
        self.assertEqual(Origin.get_tag(data[u'oed']), Origin(path=file_name, line_num=3, col_num=22))

    def test_parse_list(self):
        stream = StringIO(u"""
                - a
                - b
                """)
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, [u'a', u'b'])
        self.assertEqual(len(data), 2)
        self.assertIsInstance(data[0], str)

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=2, col_num=17))

        self.assertEqual(Origin.get_tag(data[0]), Origin(path=file_name, line_num=2, col_num=19))
        self.assertEqual(Origin.get_tag(data[1]), Origin(path=file_name, line_num=3, col_num=19))

    def test_parse_short_dict(self):
        stream = StringIO(u"""{"foo": "bar"}""")
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, dict(foo=u'bar'))

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=1, col_num=1))
        self.assertEqual(Origin.get_tag(data[u'foo']), Origin(path=file_name, line_num=1, col_num=9))

        stream = StringIO(u"""foo: bar""")
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, dict(foo=u'bar'))

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=1, col_num=1))
        self.assertEqual(Origin.get_tag(data[u'foo']), Origin(path=file_name, line_num=1, col_num=6))

    def test_error_conditions(self):
        stream = StringIO(u"""{""")
        stream.name = file_name
        with self.assertRaises(ParserError):
            yaml.load(stream, Loader=AnsibleLoader)

    def test_tab_error(self):
        stream = StringIO(u"""---\nhosts: localhost\nvars:\n  foo: bar\n\tblip: baz""")
        stream.name = file_name
        with self.assertRaises(ScannerError):
            yaml.load(stream, Loader=AnsibleLoader)

    def test_front_matter(self):
        stream = StringIO(u"""---\nfoo: bar""")
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, dict(foo=u'bar'))

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=2, col_num=1))
        self.assertEqual(Origin.get_tag(data[u'foo']), Origin(path=file_name, line_num=2, col_num=6))

        # Initial indent (See: #6348)
        stream = StringIO(u""" - foo: bar\n   baz: qux""")
        stream.name = file_name
        data = yaml.load(stream, Loader=AnsibleLoader)
        self.assertEqual(data, [{u'foo': u'bar', u'baz': u'qux'}])

        self.assertEqual(Origin.get_tag(data), Origin(path=file_name, line_num=1, col_num=2))
        self.assertEqual(Origin.get_tag(data[0]), Origin(path=file_name, line_num=1, col_num=4))
        self.assertEqual(Origin.get_tag(data[0][u'foo']), Origin(path=file_name, line_num=1, col_num=9))
        self.assertEqual(Origin.get_tag(data[0][u'baz']), Origin(path=file_name, line_num=2, col_num=9))


class TestAnsibleLoaderVault(unittest.TestCase, YamlTestUtils):
    def setUp(self):
        self.vault_password = "hunter42"
        vault_secret = TextVaultSecret(self.vault_password)
        self.vault_secrets = [('vault_secret', vault_secret),
                              ('default', vault_secret)]
        self.vault = vault.VaultLib(self.vault_secrets)

    @property
    def vault_secret(self):
        return vault.match_encrypt_secret(self.vault_secrets)[1]

    def test_wrong_password(self):
        plaintext = u"Ansible"
        bob_password = "this is a different password"

        bobs_secret = TextVaultSecret(bob_password)
        bobs_secrets = [('default', bobs_secret)]

        bobs_vault = vault.VaultLib(bobs_secrets)

        ciphertext = bobs_vault.encrypt(plaintext, vault.match_encrypt_secret(bobs_secrets)[1])

        try:
            self.vault.decrypt(ciphertext)
        except Exception as e:
            self.assertIsInstance(e, errors.AnsibleError)
            self.assertEqual(e.message, 'Decryption failed (no vault secrets were found that could decrypt).')

    def _encrypt_plaintext(self, plaintext):
        # Construct a yaml repr of a vault by hand
        vaulted_var_bytes = self.vault.encrypt(plaintext, self.vault_secret)

        # add yaml tag
        vaulted_var = vaulted_var_bytes.decode()
        lines = vaulted_var.splitlines()
        lines2 = []
        for line in lines:
            lines2.append('        %s' % line)

        vaulted_var = '\n'.join(lines2)
        tagged_vaulted_var = u"""!vault |\n%s""" % vaulted_var
        return tagged_vaulted_var

    def _build_stream(self, yaml_text):
        stream = StringIO(yaml_text)
        stream.name = 'my.yml'
        return stream

    def _load_yaml(self, yaml_text, password):
        stream = self._build_stream(yaml_text)
        data = yaml.load(stream, Loader=AnsibleLoader)

        return data


class TestAnsibleLoaderPlay(unittest.TestCase):

    def setUp(self):
        stream = StringIO("""
                - hosts: localhost
                  vars:
                    number: 1
                    string: Ansible
                    utf8_string: Cafè Eñyei
                    dictionary:
                      webster: daniel
                      oed: oxford
                    list:
                      - a
                      - b
                      - 1
                      - 2
                  tasks:
                    - name: Test case
                      ping:
                        data: "{{ utf8_string }}"

                    - name: Test 2
                      ping:
                        data: "Cafè Eñyei"

                    - name: Test 3
                      command: "printf 'Cafè Eñyei\\n'"
                      vars:
                        not_safe: !unsafe "{{ sorry }}"
                        also_not_safe: !unsafe ["{{ sorry }}"]
                """)
        self.play_filename = '/path/to/myplay.yml'
        stream.name = self.play_filename
        self.data = yaml.load(stream, Loader=AnsibleLoader)

    def tearDown(self):
        pass

    def test_data_complete(self):
        self.assertEqual(len(self.data), 1)
        self.assertIsInstance(self.data, list)
        self.assertEqual(frozenset(self.data[0].keys()), frozenset((u'hosts', u'vars', u'tasks')))

        self.assertEqual(self.data[0][u'hosts'], u'localhost')

        self.assertEqual(self.data[0][u'vars'][u'number'], 1)
        self.assertEqual(self.data[0][u'vars'][u'string'], u'Ansible')
        self.assertEqual(self.data[0][u'vars'][u'utf8_string'], u'Cafè Eñyei')
        self.assertEqual(self.data[0][u'vars'][u'dictionary'], {
            u'webster': u'daniel',
            u'oed': u'oxford'
        })
        self.assertEqual(self.data[0][u'vars'][u'list'], [u'a', u'b', 1, 2])

        self.assertEqual(self.data[0][u'tasks'], [
            {u'name': u'Test case', u'ping': {u'data': u'{{ utf8_string }}'}},
            {u'name': u'Test 2', u'ping': {u'data': u'Cafè Eñyei'}},
            {u'name': u'Test 3', u'command': u'printf \'Cafè Eñyei\n\'', 'vars': {'not_safe': "{{ sorry }}", 'also_not_safe': ["{{ sorry }}"]}},
        ])

    def walk(self, data):
        # Make sure there's no str in the data
        self.assertNotIsInstance(data, bytes)

        # Descend into various container types
        if isinstance(data, str):
            # strings are a sequence so we have to be explicit here
            return
        elif isinstance(data, (Sequence, Set)):
            for element in data:
                self.walk(element)
        elif isinstance(data, Mapping):
            for k, v in data.items():
                self.walk(k)
                self.walk(v)

        # Scalars were all checked so we're good to go
        return

    def test_no_str_in_data(self):
        # Checks that no strings are str type
        self.walk(self.data)

    def check_vars(self):
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'number']), Origin(path=self.play_filename, line_num=4, col_num=29))

        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'string']), Origin(path=self.play_filename, line_num=5, col_num=29))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'utf8_string']), Origin(path=self.play_filename, line_num=6, col_num=34))

        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'dictionary']), Origin(path=self.play_filename, line_num=8, col_num=23))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'dictionary'][u'webster']),
                         Origin(path=self.play_filename, line_num=8, col_num=32))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'dictionary'][u'oed']),
                         Origin(path=self.play_filename, line_num=9, col_num=28))

        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'list']), Origin(path=self.play_filename, line_num=11, col_num=23))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'list'][0]), Origin(path=self.play_filename, line_num=11, col_num=25))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'list'][1]), Origin(path=self.play_filename, line_num=12, col_num=25))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'list'][2]), Origin(path=self.play_filename, line_num=13, col_num=25))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars'][u'list'][3]), Origin(path=self.play_filename, line_num=14, col_num=25))

    def check_tasks(self):
        #
        # First Task
        #
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][0]), Origin(path=self.play_filename, line_num=16, col_num=23))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][0][u'name']), Origin(path=self.play_filename, line_num=16, col_num=29))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][0][u'ping']), Origin(path=self.play_filename, line_num=18, col_num=25))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][0][u'ping'][u'data']),
                         Origin(path=self.play_filename, line_num=18, col_num=31))

        #
        # Second Task
        #
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][1]), Origin(path=self.play_filename, line_num=20, col_num=23))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][1][u'name']), Origin(path=self.play_filename, line_num=20, col_num=29))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][1][u'ping']), Origin(path=self.play_filename, line_num=22, col_num=25))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][1][u'ping'][u'data']),
                         Origin(path=self.play_filename, line_num=22, col_num=31))

        #
        # Third Task
        #
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][2]), Origin(path=self.play_filename, line_num=24, col_num=23))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][2][u'name']), Origin(path=self.play_filename, line_num=24, col_num=29))
        self.assertEqual(Origin.get_tag(self.data[0][u'tasks'][2][u'command']), Origin(path=self.play_filename, line_num=25, col_num=32))

        not_safe = self.data[0][u'tasks'][2][u'vars']['not_safe']
        also_not_safe = self.data[0][u'tasks'][2][u'vars']['also_not_safe']

        assert Origin.get_tag(not_safe) == Origin(path=self.play_filename, line_num=27, col_num=35)
        assert Origin.get_tag(also_not_safe) == Origin(path=self.play_filename, line_num=28, col_num=40)
        assert Origin.get_tag(also_not_safe) == Origin(path=self.play_filename, line_num=28, col_num=40)

        assert not TrustedAsTemplate.is_tagged_on(not_safe)
        assert not TrustedAsTemplate.is_tagged_on(also_not_safe)
        assert not TrustedAsTemplate.is_tagged_on(also_not_safe[0])

    def test_line_numbers(self):
        # Check the line/column numbers are correct
        # Note: Remember, currently dicts begin at the start of their first entry
        self.assertEqual(Origin.get_tag(self.data[0]), Origin(path=self.play_filename, line_num=2, col_num=19))
        self.assertEqual(Origin.get_tag(self.data[0][u'hosts']), Origin(path=self.play_filename, line_num=2, col_num=26))
        self.assertEqual(Origin.get_tag(self.data[0][u'vars']), Origin(path=self.play_filename, line_num=4, col_num=21))

        self.check_vars()

        self.assertEqual(Origin.get_tag(self.data[0][u'tasks']), Origin(path=self.play_filename, line_num=16, col_num=21))

        self.check_tasks()


@pytest.mark.parametrize("value", (
    "[1]",
    "{a: 1}",
    "1",
    "1.1",
    "true",
    "",
    "~",
))
def test_vault_tag_type_validation(value: str) -> None:
    with pytest.raises(AnsibleYAMLParserError) as error:
        from_yaml(f"!vault {value}")

    assert "requires a string value" in str(error.value)


class CustomMapping(c.Mapping):
    def __init__(self, data: dict) -> None:
        self._data = data

    def __getitem__(self, __key):
        return self._data[__key]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return f'{type(self).__name__}({self._data!r})'


YAML_STRINGS_AND_VALUES = (
    ("test", "test"),
    ("123", 123),
    ("1.234", 1.234),
    ("true", True),
    ("false", False),
    ("{foo: bar}", dict(foo="bar")),
    ("[1, 2, 3]", [1, 2, 3]),
    ("!!set {bar: null, foo: null}", {"foo", "bar"}),
    ("2024-01-01", datetime.date(2024, 1, 1)),
    ("2024-01-01 12:01:01", datetime.datetime(2024, 1, 1, 12, 1, 1)),
    ("null", None),
    ("!!binary |\n  aGVsbG8=", b'hello'),
)
"""These values can be round-tripped through YAML."""

YAML_STRINGS_FROM_VALUES = (
    ("{a: 1}", CustomMapping(dict(a=1))),
)
"""These values can be dumped to YAML, but are converted to another type in the process, so they can't be loaded as the original type that was dumped."""


@pytest.mark.parametrize("value, expected", YAML_STRINGS_AND_VALUES)
def test_load_data_types(value: str, expected: t.Any) -> None:
    """Verify supported data types can be YAML loaded."""
    value = Origin(path=file_name).tag(value)
    result = from_yaml(value)
    origin = Origin.get_tag(result)

    assert result == expected

    if type(result) in _untaggable_types:
        assert origin is None
    else:
        assert origin.path == file_name


@pytest.mark.parametrize("expected, value", YAML_STRINGS_AND_VALUES + YAML_STRINGS_FROM_VALUES)
def test_dump_data_types(value: str, expected: t.Any) -> None:
    """Verify supported data types can be YAML dumped."""
    result = yaml.dump(value, Dumper=AnsibleDumper, default_flow_style=True).rstrip()

    assert result == expected


@pytest.mark.parametrize("trust_input_str", (
    True,
    False,
))
def test_string_trust_propagation(trust_input_str: bool) -> None:
    """
    Verify that input trust propagation behaves as expected. An explicit boolean `trusted_as_template` arg to the loader is always
    respected; if not specified, the presence of trust on the input string determines if trust is applied to outputs.
    """
    data = "foo: bar"

    if trust_input_str:
        data = TrustedAsTemplate().tag(data)

    res = yaml.load(data, Loader=AnsibleLoader)  # type: ignore[arg-type]

    assert trust_input_str == TrustedAsTemplate.is_tagged_on(res['foo'])



# Copyright (c) 2018, Ansible Project
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)
"""
Modules in _utils are waiting to find a better home.  If you need to use them, be prepared for them
to move to a different location in the future.
"""

from __future__ import annotations

import inspect
import typing as t

_Type = t.TypeVar('_Type')


def get_all_subclasses(cls: type[_Type], *, include_abstract: bool = True, consider_self: bool = False) -> set[type[_Type]]:
    """Recursively find all subclasses of a given type, including abstract classes by default."""
    subclasses: set[type[_Type]] = {cls} if consider_self else set()
    queue: list[type[_Type]] = [cls]

    while queue:
        parent = queue.pop()

        for child in parent.__subclasses__():
            if child in subclasses:
                continue

            queue.append(child)
            subclasses.add(child)

    if not include_abstract:
        subclasses = {sc for sc in subclasses if not inspect.isabstract(sc)}

    return subclasses



# Copyright 2018, Matt Martz <matt@sivel.net>
# Copyright 2019, Andrew Klychkov @Andersson007 <aaklychkov@mail.ru>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os
import json
import pathlib

import pytest

from collections.abc import Mapping
from datetime import date, datetime, timezone, timedelta

from ansible._internal._json._legacy_encoder import LegacyControllerJSONEncoder
from ansible.module_utils.common.json import get_encoder, get_decoder
from ansible.parsing.vault import EncryptedString, AnsibleVaultError
from ansible._internal._datatag._tags import TrustedAsTemplate
from ansible._internal._json._profiles import _legacy


def test_AnsibleJSONDecoder_vault(_empty_vault_secrets_context):
    profile = _legacy

    with open(os.path.join(os.path.dirname(__file__), 'fixtures/ajson.json')) as f:
        data = json.load(f, cls=get_decoder(profile))

    assert isinstance(data['password'], EncryptedString)
    with pytest.raises(AnsibleVaultError):
        str(data['password'])

    assert isinstance(data['bar']['baz'][0]['password'], EncryptedString)
    with pytest.raises(AnsibleVaultError):
        str(data['bar']['baz'][0]['password'])

    assert isinstance(data['foo']['password'], EncryptedString)
    with pytest.raises(AnsibleVaultError):
        str(data['foo']['password'])


def vault_data():
    """
    Prepare vault test data for AnsibleJSONEncoder.default().

    Return a list of tuples (input, expected).
    """

    profile = _legacy

    raw_data = TrustedAsTemplate().tag((pathlib.Path(__file__).parent / 'fixtures/ajson.json').read_text())
    data = json.loads(raw_data, cls=get_decoder(profile))

    data_0 = data['password']
    data_1 = data['bar']['baz'][0]['password']

    expected_0 = (u'$ANSIBLE_VAULT;1.1;AES256\n34646264306632313333393636316'
                  '562356435376162633631326264383934326565333633366238\n3863'
                  '373264326461623132613931346165636465346337310a32643431383'
                  '0316337393263616439\n646539373134633963666338613632666334'
                  '65663730303633323534363331316164623237363831\n35363335613'
                  '93238370a313330316263373938326162386433313336613532653538'
                  '376662306435\n3339\n')

    expected_1 = (u'$ANSIBLE_VAULT;1.1;AES256\n34646264306632313333393636316'
                  '562356435376162633631326264383934326565333633366238\n3863'
                  '373264326461623132613931346165636465346337310a32643431383'
                  '0316337393263616439\n646539373134633963666338613632666334'
                  '65663730303633323534363331316164623237363831\n35363335613'
                  '93238370a313330316263373938326162386433313336613532653538'
                  '376662306435\n3338\n')

    return [
        (data_0, expected_0),
        (data_1, expected_1),
    ]


class TestAnsibleJSONEncoder:

    """
    Namespace for testing AnsibleJSONEncoder.
    """

    @pytest.fixture(scope='class')
    def mapping(self, request):
        """
        Returns object of Mapping mock class.

        The object is used for testing handling of Mapping objects
        in AnsibleJSONEncoder.default().
        Using a plain dictionary instead is not suitable because
        it is handled by default encoder of the superclass (json.JSONEncoder).
        """

        class M(Mapping):

            """Mock mapping class."""

            def __init__(self, *args, **kwargs):
                self.__dict__.update(*args, **kwargs)

            def __getitem__(self, key):
                return self.__dict__[key]

            def __iter__(self):
                return iter(self.__dict__)

            def __len__(self):
                return len(self.__dict__)

        mapping = M(request.param)

        assert isinstance(len(mapping), int)   # ensure coverage of __len__

        return mapping

    @pytest.fixture
    def ansible_json_encoder(self):
        """Return AnsibleJSONEncoder object."""
        return LegacyControllerJSONEncoder()

    ###############
    # Test methods:

    @pytest.mark.parametrize(
        'test_input,expected',
        [
            (datetime(2019, 5, 14, 13, 39, 38, 569047), '2019-05-14T13:39:38.569047'),
            (datetime(2019, 5, 14, 13, 47, 16, 923866), '2019-05-14T13:47:16.923866'),
            (date(2019, 5, 14), '2019-05-14'),
            (date(2020, 5, 14), '2020-05-14'),
            (datetime(2019, 6, 15, 14, 45, tzinfo=timezone.utc), '2019-06-15T14:45:00+00:00'),
            (datetime(2019, 6, 15, 14, 45, tzinfo=timezone(timedelta(hours=1, minutes=40))), '2019-06-15T14:45:00+01:40'),
        ]
    )
    def test_date_datetime(self, ansible_json_encoder, test_input, expected):
        """
        Test for passing datetime.date or datetime.datetime objects to AnsibleJSONEncoder.default().
        """
        assert ansible_json_encoder.default(test_input) == expected

    @pytest.mark.parametrize(
        'mapping,expected',
        [
            ({'1': 1}, {'1': 1}),
            ({'2': 2}, {'2': 2}),
            ({'1': 2}, {'1': 2}),
            ({'2': 1}, {'2': 1}),
        ], indirect=['mapping'],
    )
    def test_mapping(self, ansible_json_encoder, mapping, expected):
        """
        Test for passing Mapping object to AnsibleJSONEncoder.default().
        """
        assert ansible_json_encoder.default(mapping) == expected

    @pytest.mark.parametrize('test_input,expected', vault_data())
    def test_ansible_json_encoder_vault(self, test_input, expected):
        """
        Test for passing vaulted values to AnsibleJSONEncoder.default().
        """
        profile = _legacy
        assert json.dumps(test_input, cls=get_encoder(profile)) == '{"__ansible_vault": "%s"}' % expected.replace('\n', '\\n')

    @pytest.mark.parametrize(
        'test_input,expected',
        [
            ({'1': 'first'}, {'1': 'first'}),
            ({'2': 'second'}, {'2': 'second'}),
        ]
    )
    def test_default_encoder(self, ansible_json_encoder, test_input, expected):
        """
        Test for the default encoder of AnsibleJSONEncoder.default().

        If objects of different classes that are not tested above were passed,
        AnsibleJSONEncoder.default() invokes 'default()' method of json.JSONEncoder superclass.
        """
        assert ansible_json_encoder.default(test_input) == expected


@pytest.mark.parametrize("trust_input_str", (
    True,
    False
))
def test_string_trust_propagation(trust_input_str: bool) -> None:
    """Verify that input trust propagation behaves as expected. The presence of trust on the input string determines if trust is applied to outputs."""
    data = '{"foo": "bar"}'

    if trust_input_str:
        data = TrustedAsTemplate().tag(data)

    res = json.loads(data, cls=_legacy.Decoder)

    assert trust_input_str == TrustedAsTemplate.is_tagged_on(res['foo'])



from __future__ import annotations

from ansible.module_utils.basic import AnsibleModule


def main():
    module = AnsibleModule(
        argument_spec=dict(
            data=dict(type='raw'),
        ),
        supports_check_mode=True,
    )

    if module.params['data'] == 'crash':
        raise Exception("boom")

    result = dict(
        data=module.params['data'],
    )

    module.exit_json(**result)


if __name__ == '__main__':
    main()



#
# (c) 2018 Red Hat Inc.
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import annotations


import os
import time
import re

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_text, to_bytes
from ansible.module_utils.six.moves.urllib.parse import urlsplit
from ansible.plugins.action.normal import ActionModule as _ActionModule
from ansible.utils.display import Display

display = Display()

PRIVATE_KEYS_RE = re.compile("__.+__")


class ActionModule(_ActionModule):
    def run(self, task_vars=None):
        config_module = hasattr(self, "_config_module") and self._config_module
        if config_module and self._task.args.get("src"):
            try:
                self._handle_src_option()
            except AnsibleError as e:
                return {"failed": True, "msg": e.message, "changed": False}

        result = super(ActionModule, self).run(task_vars=task_vars)

        if (
            config_module
            and self._task.args.get("backup")
            and not result.get("failed")
        ):
            self._handle_backup_option(result, task_vars)

        return result

    def _handle_backup_option(self, result, task_vars):

        filename = None
        backup_path = None
        try:
            content = result["__backup__"]
        except KeyError:
            raise AnsibleError("Failed while reading configuration backup")

        backup_options = self._task.args.get("backup_options")
        if backup_options:
            filename = backup_options.get("filename")
            backup_path = backup_options.get("dir_path")

        if not backup_path:
            cwd = self._get_working_path()
            backup_path = os.path.join(cwd, "backup")
        if not filename:
            tstamp = time.strftime(
                "%Y-%m-%d@%H:%M:%S", time.localtime(time.time())
            )
            filename = "%s_config.%s" % (
                task_vars["inventory_hostname"],
                tstamp,
            )

        dest = os.path.join(backup_path, filename)
        backup_path = os.path.expanduser(
            os.path.expandvars(
                to_bytes(backup_path, errors="surrogate_or_strict")
            )
        )

        if not os.path.exists(backup_path):
            os.makedirs(backup_path)

        new_task = self._task.copy()
        for item in self._task.args:
            if not item.startswith("_"):
                new_task.args.pop(item, None)

        new_task.args.update(dict(content=content, dest=dest))
        copy_action = self._shared_loader_obj.action_loader.get(
            "copy",
            task=new_task,
            connection=self._connection,
            play_context=self._play_context,
            loader=self._loader,
            templar=self._templar,
            shared_loader_obj=self._shared_loader_obj,
        )
        copy_result = copy_action.run(task_vars=task_vars)
        if copy_result.get("failed"):
            result["failed"] = copy_result["failed"]
            result["msg"] = copy_result.get("msg")
            return

        result["backup_path"] = dest
        if copy_result.get("changed", False):
            result["changed"] = copy_result["changed"]

        if backup_options and backup_options.get("filename"):
            result["date"] = time.strftime(
                "%Y-%m-%d",
                time.gmtime(os.stat(result["backup_path"]).st_ctime),
            )
            result["time"] = time.strftime(
                "%H:%M:%S",
                time.gmtime(os.stat(result["backup_path"]).st_ctime),
            )

        else:
            result["date"] = tstamp.split("@")[0]
            result["time"] = tstamp.split("@")[1]
            result["shortname"] = result["backup_path"][::-1].split(".", 1)[1][
                ::-1
            ]
            result["filename"] = result["backup_path"].split("/")[-1]

        # strip out any keys that have two leading and two trailing
        # underscore characters
        for key in list(result.keys()):
            if PRIVATE_KEYS_RE.match(key):
                del result[key]

    def _get_working_path(self):
        cwd = self._loader.get_basedir()
        if self._task._role is not None:
            cwd = self._task._role._role_path
        return cwd

    def _handle_src_option(self, convert_data=True):
        src = self._task.args.get("src")
        working_path = self._get_working_path()

        if os.path.isabs(src) or urlsplit("src").scheme:
            source = src
        else:
            source = self._loader.path_dwim_relative(
                working_path, "templates", src
            )
            if not source:
                source = self._loader.path_dwim_relative(working_path, src)

        if not os.path.exists(source):
            raise AnsibleError("path specified in src not found")

        try:
            with open(source, "r") as f:
                template_data = to_text(f.read())
        except IOError as e:
            raise AnsibleError(
                "unable to load src file {0}, I/O error({1}): {2}".format(
                    source, e.errno, e.strerror
                )
            )

        # Create a template search path in the following order:
        # [working_path, self_role_path, dependent_role_paths, dirname(source)]
        searchpath = [working_path]
        if self._task._role is not None:
            searchpath.append(self._task._role._role_path)
            if hasattr(self._task, "_block:"):
                dep_chain = self._task._block.get_dep_chain()
                if dep_chain is not None:
                    for role in dep_chain:
                        searchpath.append(role._role_path)
        searchpath.append(os.path.dirname(source))
        templar = self._templar.copy_with_new_env(searchpath=searchpath)
        self._task.args["src"] = templar.template(template_data)

    def _get_network_os(self, task_vars):
        if "network_os" in self._task.args and self._task.args["network_os"]:
            display.vvvv("Getting network OS from task argument")
            network_os = self._task.args["network_os"]
        elif self._play_context.network_os:
            display.vvvv("Getting network OS from inventory")
            network_os = self._play_context.network_os
        elif (
            "network_os" in task_vars.get("ansible_facts", {})
            and task_vars["ansible_facts"]["network_os"]
        ):
            display.vvvv("Getting network OS from fact")
            network_os = task_vars["ansible_facts"]["network_os"]
        else:
            raise AnsibleError(
                "ansible_network_os must be specified on this host"
            )

        return network_os



# (c) 2012-2014, Michael DeHaan <michael.dehaan@gmail.com>
# (c) 2015, Toshio Kuraotmi <tkuratomi@ansible.com>
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from collections import defaultdict

from unittest import mock

import unittest

from ansible.errors import AnsibleError
from ansible._internal._datatag._tags import Origin
from ansible.parsing.vault import EncryptedString
from ansible.utils.vars import combine_vars, merge_hash, transform_to_native_types
from units.mock.vault_helper import VaultTestHelper


class TestVariableUtils(unittest.TestCase):
    combine_vars_merge_data = (
        dict(
            a=dict(a=1),
            b=dict(b=2),
            result=dict(a=1, b=2),
        ),
        dict(
            a=dict(a=1, c=dict(foo='bar')),
            b=dict(b=2, c=dict(baz='bam')),
            result=dict(a=1, b=2, c=dict(foo='bar', baz='bam'))
        ),
        dict(
            a=defaultdict(a=1, c=defaultdict(foo='bar')),
            b=dict(b=2, c=dict(baz='bam')),
            result=defaultdict(a=1, b=2, c=defaultdict(foo='bar', baz='bam'))
        ),
    )
    combine_vars_replace_data = (
        dict(
            a=dict(a=1),
            b=dict(b=2),
            result=dict(a=1, b=2)
        ),
        dict(
            a=dict(a=1, c=dict(foo='bar')),
            b=dict(b=2, c=dict(baz='bam')),
            result=dict(a=1, b=2, c=dict(baz='bam'))
        ),
        dict(
            a=defaultdict(a=1, c=dict(foo='bar')),
            b=dict(b=2, c=defaultdict(baz='bam')),
            result=defaultdict(a=1, b=2, c=defaultdict(baz='bam'))
        ),
    )

    def test_combine_vars_improper_args(self):
        with mock.patch('ansible.constants.DEFAULT_HASH_BEHAVIOUR', 'replace'):
            with self.assertRaises(AnsibleError):
                combine_vars([1, 2, 3], dict(a=1))
            with self.assertRaises(AnsibleError):
                combine_vars(dict(a=1), [1, 2, 3])

        with mock.patch('ansible.constants.DEFAULT_HASH_BEHAVIOUR', 'merge'):
            with self.assertRaises(AnsibleError):
                combine_vars([1, 2, 3], dict(a=1))
            with self.assertRaises(AnsibleError):
                combine_vars(dict(a=1), [1, 2, 3])

    def test_combine_vars_replace(self):
        with mock.patch('ansible.constants.DEFAULT_HASH_BEHAVIOUR', 'replace'):
            for test in self.combine_vars_replace_data:
                self.assertEqual(combine_vars(test['a'], test['b']), test['result'])

    def test_combine_vars_merge(self):
        with mock.patch('ansible.constants.DEFAULT_HASH_BEHAVIOUR', 'merge'):
            for test in self.combine_vars_merge_data:
                self.assertEqual(combine_vars(test['a'], test['b']), test['result'])

    merge_hash_data = {
        "low_prio": {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "low_value",
                    "list": ["low_value"]
                }
            },
            "b": [1, 1, 2, 3]
        },
        "high_prio": {
            "a": {
                "a'": {
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["high_value"]
                }
            },
            "b": [3, 4, 4, {"5": "value"}]
        }
    }

    def test_merge_hash_simple(self):
        for test in self.combine_vars_merge_data:
            self.assertEqual(merge_hash(test['a'], test['b']), test['result'])

        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["high_value"]
                }
            },
            "b": high['b']
        }
        self.assertEqual(merge_hash(low, high), expected)

    def test_merge_hash_non_recursive_and_list_replace(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = high
        self.assertEqual(merge_hash(low, high, False, 'replace'), expected)

    def test_merge_hash_non_recursive_and_list_keep(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": high['a'],
            "b": low['b']
        }
        self.assertEqual(merge_hash(low, high, False, 'keep'), expected)

    def test_merge_hash_non_recursive_and_list_append(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": high['a'],
            "b": low['b'] + high['b']
        }
        self.assertEqual(merge_hash(low, high, False, 'append'), expected)

    def test_merge_hash_non_recursive_and_list_prepend(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": high['a'],
            "b": high['b'] + low['b']
        }
        self.assertEqual(merge_hash(low, high, False, 'prepend'), expected)

    def test_merge_hash_non_recursive_and_list_append_rp(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": high['a'],
            "b": [1, 1, 2] + high['b']
        }
        self.assertEqual(merge_hash(low, high, False, 'append_rp'), expected)

    def test_merge_hash_non_recursive_and_list_prepend_rp(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": high['a'],
            "b": high['b'] + [1, 1, 2]
        }
        self.assertEqual(merge_hash(low, high, False, 'prepend_rp'), expected)

    def test_merge_hash_recursive_and_list_replace(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["high_value"]
                }
            },
            "b": high['b']
        }
        self.assertEqual(merge_hash(low, high, True, 'replace'), expected)

    def test_merge_hash_recursive_and_list_keep(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["low_value"]
                }
            },
            "b": low['b']
        }
        self.assertEqual(merge_hash(low, high, True, 'keep'), expected)

    def test_merge_hash_recursive_and_list_append(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["low_value", "high_value"]
                }
            },
            "b": low['b'] + high['b']
        }
        self.assertEqual(merge_hash(low, high, True, 'append'), expected)

    def test_merge_hash_recursive_and_list_prepend(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["high_value", "low_value"]
                }
            },
            "b": high['b'] + low['b']
        }
        self.assertEqual(merge_hash(low, high, True, 'prepend'), expected)

    def test_merge_hash_recursive_and_list_append_rp(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["low_value", "high_value"]
                }
            },
            "b": [1, 1, 2] + high['b']
        }
        self.assertEqual(merge_hash(low, high, True, 'append_rp'), expected)

    def test_merge_hash_recursive_and_list_prepend_rp(self):
        low = self.merge_hash_data['low_prio']
        high = self.merge_hash_data['high_prio']
        expected = {
            "a": {
                "a'": {
                    "x": "low_value",
                    "y": "high_value",
                    "z": "high_value",
                    "list": ["high_value", "low_value"]
                }
            },
            "b": high['b'] + [1, 1, 2]
        }
        self.assertEqual(merge_hash(low, high, True, 'prepend_rp'), expected)


def test_transform_to_native_types() -> None:
    """Verify that transform_to_native_types results in native types for both keys and values, with default redaction."""
    value = {
        Origin(description="blah").tag("tagged_key"): Origin(description="blah").tag("value with tagged key"),
        # use a bogus EncryptedString instance with no VaultSecretsContext active; ensures that transform with redaction does not attempt decryption
        "redact_this": EncryptedString(ciphertext="bogus")
    }

    result = transform_to_native_types(value)

    assert result == dict(tagged_key="value with tagged key", redact_this='<redacted>')

    assert all(type(key) is str for key in result.keys())  # pylint: disable=unidiomatic-typecheck
    assert all(type(value) is str for value in result.values())  # pylint: disable=unidiomatic-typecheck


def test_transform_to_native_types_unredacted(_vault_secrets_context: None) -> None:
    """Verify that transform with redaction disabled returns a plain string decrypted value."""
    plaintext = "hello"
    value = dict(enc=VaultTestHelper.make_encrypted_string(plaintext))
    result = transform_to_native_types(value, redact=False)

    assert result == dict(enc=plaintext)
    assert all(type(key) is str for key in result.keys())  # pylint: disable=unidiomatic-typecheck
    assert all(type(value) is str for value in result.values())  # pylint: disable=unidiomatic-typecheck



from __future__ import annotations

import copy
import pickle
import typing as t

import pytest

from ansible.module_utils._internal._datatag import AnsibleTagHelper
from ansible.parsing.vault import EncryptedString, VaultSecretsContext, VaultSecret, VaultLib
from ansible._internal._datatag._tags import Origin, TrustedAsTemplate, VaultedValue
from ..module_utils.datatag.test_datatag import TestDatatagTarget as _TestDatatagTarget, Later


def _get_test_secret():
    return VaultSecret(b'secretbytesblah')


@pytest.fixture(scope='module')
def _vault_secrets_context() -> t.Generator[None]:
    """A fixture that provides a `VaultSecretsContext` populated with a single default secret under the default id."""
    VaultSecretsContext._current = None

    secret = _get_test_secret()

    VaultSecretsContext.initialize(VaultSecretsContext(secrets=[('default', secret)]))

    try:
        yield
    finally:
        VaultSecretsContext._current = None


pytestmark = pytest.mark.usefixtures('_vault_secrets_context')


class TestDatatagController(_TestDatatagTarget):
    later = t.cast(t.Self, Later(locals(), parent_type=_TestDatatagTarget))

    tag_instances_with_reprs = [
        (Origin(path='/himom.yml', line_num=42, col_num=42), "Origin(path='/himom.yml', line_num=42, col_num=42)"),
        (TrustedAsTemplate(), "TrustedAsTemplate()"),
        (VaultedValue(ciphertext="hi mom I am a secret"), "VaultedValue(ciphertext='hi mom I am a secret')"),
    ]

    test_dataclass_tag_base_field_validation_fail_instances = [
        (Origin, dict(path=TrustedAsTemplate().tag(''))),
        (Origin, dict(line_num=TrustedAsTemplate().tag(1), path='')),
        (Origin, dict(col_num=TrustedAsTemplate().tag(1), path='')),
        (VaultedValue, dict(ciphertext=TrustedAsTemplate().tag(''))),
    ]

    test_dataclass_tag_base_field_validation_pass_instances = [
        (Origin, dict(path='/something')),
        (Origin, dict(path='/something', line_num=1)),
        (Origin, dict(path='/something', col_num=1)),
        (VaultedValue, dict(ciphertext='')),
    ]

    @classmethod
    def post_init(cls) -> None:
        ciphertext = VaultLib(None).encrypt("i am a secret", _get_test_secret()).decode()

        cls.taggable_instances += [
            EncryptedString(ciphertext=ciphertext),
        ]

    # DTFIX5: ensure we're calculating the correct set of values for this context
    @classmethod
    def container_test_cases(cls) -> list:
        return []

    # HACK: avoid `SKIPPED` notifications for inherited tests with no data
    def test_tag_copy(self) -> None:
        pass

    def test_instance_copy_roundtrip(self) -> None:
        pass

    def test_tag(self) -> None:
        pass

    @pytest.mark.autoparam(later.serializable_instances)
    def test_deepcopy_roundtrip(self, value: object):
        super().test_deepcopy_roundtrip(value)


@pytest.mark.parametrize("sp, value", (
    (Origin(path="/hi"), "/hi"),
    (Origin(path="/hi", line_num=1), "/hi:1"),
    (Origin(path="/hi", line_num=1, col_num=2), "/hi:1:2"),
    (Origin(path="/hi", col_num=2), "/hi"),
    (Origin(path="/hi", line_num=0), "/hi"),
    (Origin(path="/hi", line_num=0, col_num=0), "/hi"),
    (Origin(path="/hi", col_num=0), "/hi"),
    (Origin(path="/hi", line_num=-1), "/hi"),
    (Origin(path="/hi", line_num=1, col_num=-1), "/hi:1"),
    (Origin(description='<something>'), "<something>"),
    (Origin(description='<something>', line_num=1), "<something>:1"),
    (Origin(path="/hi", description='<something>'), "/hi (<something>)"),
    (Origin(path="/hi", description='<something>', line_num=1), "/hi:1 (<something>)"),
), ids=str)
def test_origin_str(sp: Origin, value: str) -> None:
    assert str(sp) == value


def test_tag_builtins():
    values = [123, 123.45, 'a string value', tuple([1, 2, 3]), [1, 2, 3], {1, 2, 3}, dict(one=1, two=2)]

    for original_val in values:
        tagged_val = TrustedAsTemplate().tag(original_val)
        zero_tagged_val = AnsibleTagHelper.tag(original_val, [])  # should return original value, not an empty tagged obj

        assert original_val == tagged_val  # equality should pass
        assert not TrustedAsTemplate.is_tagged_on(original_val)  # immutable original value via bool check
        assert TrustedAsTemplate.get_tag(original_val) is None  # immutable original value via get_tag
        assert not AnsibleTagHelper.tags(original_val)  # immutable original value via tags

        assert TrustedAsTemplate.is_tagged_on(tagged_val)
        assert TrustedAsTemplate.get_tag(tagged_val) is TrustedAsTemplate()  # singleton tag type, should be reference-equal
        assert original_val is zero_tagged_val  # original value should reference-equal the zero-tagged value

        origin = Origin(path="/foo", line_num=12, col_num=34)

        multi_tagged_val = origin.tag(tagged_val)
        assert tagged_val is not multi_tagged_val
        assert TrustedAsTemplate.is_tagged_on(multi_tagged_val)
        assert Origin.is_tagged_on(multi_tagged_val)
        assert TrustedAsTemplate.get_tag(multi_tagged_val) is TrustedAsTemplate()  # singleton tag type, should be reference-equal
        assert Origin.get_tag(multi_tagged_val) is origin


# pylint: disable=unnecessary-lambda
@pytest.mark.parametrize("copy_expr", (
    lambda es: copy.copy(es),
    lambda es: copy.deepcopy(es),
    lambda es: pickle.loads(pickle.dumps(es)),
))
def test_encrypted_string_copies(copy_expr: t.Callable[[EncryptedString], EncryptedString]):
    """Validate that copy/deepcopy/pickle work with `EncryptedString`."""
    plaintext = "i am a secret"
    ciphertext = VaultLib(None).encrypt(plaintext, _get_test_secret()).decode()
    es = EncryptedString(ciphertext=ciphertext)
    copied_es = copy_expr(es)
    assert copied_es is not es
    assert copied_es == es


@pytest.mark.parametrize("comparison, expected", (
    ("==", True),
    ("!=", False),
    (">=", True),
    ("<=", True),
    (">", False),
    ("<", False),
))
def test_encrypted_string_binary_operators(comparison: str, expected: bool) -> None:
    """Validate binary operator behavior with permutations of plain strings and `EncryptedString` instances."""
    plaintext = "i am a secret"
    ciphertext = VaultLib(None).encrypt(plaintext, _get_test_secret()).decode()
    es = EncryptedString(ciphertext=ciphertext)
    copied_es = copy.copy(es)

    permutations = (
        (plaintext, plaintext),
        (plaintext, copied_es),
        (copied_es, plaintext),
        (es, copied_es),
    )

    for lhs, rhs in permutations:
        assert eval(f'{lhs!r} {comparison} {rhs!r}') == expected

        object.__setattr__(es, '_plaintext', None)
        object.__setattr__(copied_es, '_plaintext', None)
